import nengo
from nengo.exceptions import ValidationError
from nengo.utils.compat import is_array, is_integer, is_number, range
import numpy as np

from nengo_spa.ast.base import Fixed, infer_types, TypeCheckedBinaryOp
from nengo_spa.types import TAnyVocab, TScalar, TVocabulary


class SemanticPointer(Fixed):
    """A Semantic Pointer, based on Holographic Reduced Representations.

    Operators are overloaded so that ``+`` and ``-`` are addition,
    ``*`` is circular convolution, and ``~`` is the inversion operator.

    Parameters
    ----------
    data : int or array_like
        The vector constituting the Semantic Pointer. If an integer is given,
        a random unit-length vector will be generated.
    rng : numpy.random.RandomState, optional
        Random number generator used for random generation of a Semantic
        Pointer.
    vocab : Vocabulary
        Vocabulary that the Semantic Pointer is considered to be part of.

    Attributes
    ----------
    v : array_like
        The vector constituting the Semantic Pointer.
    vocab : Vocabulary or None
        The vocabulary the this Semantic Pointer is considered to be part of.
    """

    def __init__(self, data, rng=None, vocab=None):
        super(SemanticPointer, self).__init__(
            TAnyVocab if vocab is None else TVocabulary(vocab))

        if rng is None:
            rng = np.random

        if is_integer(data):
            if data < 1:
                raise ValidationError("Number of dimensions must be a "
                                      "positive int", attr='data', obj=self)

            self.v = rng.randn(data)
            self.v /= np.linalg.norm(self.v)
        else:
            self.v = np.array(data, dtype=float)
            if len(self.v.shape) != 1:
                raise ValidationError("'data' must be a vector", 'data', self)
        self.v.setflags(write=False)

        self.vocab = vocab

    def evaluate(self):
        return self

    def connect_to(self, sink, **kwargs):
        return nengo.Connection(self.construct(), sink, **kwargs)

    def construct(self):
        return nengo.Node(
            self.v, label="Semantic Pointer ({}d)".format(len(self)))

    def normalized(self):
        """Normalize the Semantic Pointer and return it as a new object.

        The original object is not modified.
        """
        nrm = np.linalg.norm(self.v)
        if nrm > 0:
            return SemanticPointer(self.v / nrm, vocab=self.vocab)

    def unitary(self):
        """Make the Semantic Pointer unitary and return it as a new object.

        The original object is not modified.

        A unitary Semantic Pointer has the property that it does not change
        the length of Semantic Pointers it is bound with using circular
        convolution.
        """
        fft_val = np.fft.fft(self.v)
        fft_imag = fft_val.imag
        fft_real = fft_val.real
        fft_norms = np.sqrt(fft_imag ** 2 + fft_real ** 2)
        fft_unit = fft_val / fft_norms
        return SemanticPointer(np.array((np.fft.ifft(
            fft_unit, n=len(self))).real), vocab=self.vocab)

    def copy(self):
        """Return another semantic pointer with the same data."""
        return SemanticPointer(data=self.v, vocab=self.vocab)

    def length(self):
        """Return the L2 norm of the vector."""
        return np.linalg.norm(self.v)

    def __len__(self):
        """Return the number of dimensions in the vector."""
        return len(self.v)

    def __str__(self):
        return str(self.v)

    @TypeCheckedBinaryOp(Fixed)
    def __add__(self, other):
        type_ = infer_types(self, other)
        vocab = None if type_ == TAnyVocab else type_.vocab
        return SemanticPointer(data=self.v + other.evaluate().v, vocab=vocab)

    def __radd__(self, other):
        return self + other

    def __neg__(self):
        return SemanticPointer(data=-self.v, vocab=self.vocab)

    @TypeCheckedBinaryOp(Fixed)
    def __sub__(self, other):
        type_ = infer_types(self, other)
        vocab = None if type_ == TAnyVocab else type_.vocab
        return SemanticPointer(data=self.v - other.evaluate().v, vocab=vocab)

    def __rsub__(self, other):
        return (-self) + other

    def __mul__(self, other):
        """Multiplication of two SemanticPointers is circular convolution.

        If multiplied by a scalar, we do normal multiplication.
        """
        if is_array(other):
            raise TypeError(
                "Multiplication of Semantic Pointers with arrays in not "
                "allowed.")
        elif is_number(other):
            return SemanticPointer(data=self.v * other, vocab=self.vocab)
        elif isinstance(other, Fixed):
            if other.type == TScalar:
                return SemanticPointer(
                    data=self.v * other.evaluate(), vocab=self.vocab)
            else:
                return self.convolve(other)
        else:
            return NotImplemented

    def __rmul__(self, other):
        """Multiplication of two SemanticPointers is circular convolution.

        If multiplied by a scalar, we do normal multiplication.
        """
        return self.__mul__(other)

    def __invert__(self):
        """Return a reorganized vector that acts as an inverse for convolution.

        This reorganization turns circular convolution into circular
        correlation, meaning that ``A*B*~B`` is approximately ``A``.

        For the vector ``[1, 2, 3, 4, 5]``, the inverse is ``[1, 5, 4, 3, 2]``.
        """
        return SemanticPointer(
            data=self.v[-np.arange(len(self))], vocab=self.vocab)

    def convolve(self, other):
        """Return the circular convolution of two SemanticPointers."""
        type_ = infer_types(self, other)
        vocab = None if type_ == TAnyVocab else type_.vocab
        n = len(self)
        x = np.fft.irfft(
            np.fft.rfft(self.v) * np.fft.rfft(other.evaluate().v), n=n)
        return SemanticPointer(data=x, vocab=vocab)

    def get_convolution_matrix(self):
        """Return the matrix that does a circular convolution by this vector.

        This should be such that
        ``A*B == dot(A.get_convolution_matrix(), B.v)``.
        """
        D = len(self.v)
        T = []
        for i in range(D):
            T.append([self.v[(i - j) % D] for j in range(D)])
        return np.array(T)

    def dot(self, other):
        """Return the dot product of the two vectors."""
        if isinstance(other, Fixed):
            infer_types(self, other)
            other = other.evaluate().v
        return np.dot(self.v, other)

    def __matmul__(self, other):
        return self.dot(other)

    def compare(self, other):
        """Return the similarity between two SemanticPointers.

        This is the normalized dot product, or (equivalently), the cosine of
        the angle between the two vectors.
        """
        if isinstance(other, SemanticPointer):
            infer_types(self, other)
            other = other.evaluate().v
        scale = np.linalg.norm(self.v) * np.linalg.norm(other)
        if scale == 0:
            return 0
        return np.dot(self.v, other) / scale

    def reinterpret(self, vocab):
        return SemanticPointer(self.v, vocab=vocab)

    def translate(self, vocab, populate=None, keys=None, solver=None):
        tr = self.vocab.transform_to(vocab, populate, solver)
        return SemanticPointer(np.dot(tr, self.evaluate().v), vocab=vocab)

    def distance(self, other):
        """Return a distance measure between the vectors.

        This is ``1-cos(angle)``, so that it is 0 when they are identical, and
        the distance gets larger as the vectors are farther apart.
        """
        return 1 - self.compare(other)

    def mse(self, other):
        """Return the mean-squared-error between two vectors."""
        if isinstance(other, SemanticPointer):
            infer_types(self, other)
            other = other.evaluate().v
        return np.sum((self.v - other)**2) / len(self.v)


class Identity(SemanticPointer):
    """Circular convolution identity.

    Parameters
    ----------
    n_dimensions : int
        Dimensionality of the identity vector.
    vocab : Vocabulary
        Vocabulary that the Semantic Pointer is considered to be part of.
    """

    def __init__(self, n_dimensions, vocab=None):
        data = np.zeros(n_dimensions)
        data[0] = 1.
        super(Identity, self).__init__(data, vocab=vocab)


class AbsorbingElement(SemanticPointer):
    r"""Circular convolution absorbing element.

    If :math:`z` denotes the absorbing element, :math:`v \circledast z = c z`,
    where :math:`v` is a Semantic Pointer and :math:`c` is a real-valued
    scalar.

    Parameters
    ----------
    n_dimensions : int
        Dimensionality of the identity vector.
    vocab : Vocabulary
        Vocabulary that the Semantic Pointer is considered to be part of.
    """
    def __init__(self, n_dimensions, vocab=None):
        data = np.ones(n_dimensions) / np.sqrt(n_dimensions)
        super(AbsorbingElement, self).__init__(data, vocab=vocab)


class Zero(SemanticPointer):
    """An all zero Semantic Pointer.

    Parameters
    ----------
    n_dimensions : int
        Dimensionality of the identity vector.
    vocab : Vocabulary
        Vocabulary that the Semantic Pointer is considered to be part of.
    """
    def __init__(self, n_dimensions, vocab=None):
        super(Zero, self).__init__(np.zeros(n_dimensions), vocab=vocab)



"""Implementation of action syntax."""

from collections import Mapping, OrderedDict

import nengo
from nengo.utils.compat import is_string
import numpy as np

from nengo_spa.ast.base import Fixed, infer_types, Node
from nengo_spa.ast import dynamic
from nengo_spa.exceptions import SpaActionSelectionError, SpaTypeError
from nengo_spa.types import TScalar


class ModuleInput(object):
    """Represents the input to a module with type information.

    Supports the ``>>`` operator to provide input from an AST node. It will
    create a simple connection by default, but a `.RoutedConnection` instance
    if used within the context of an `.ActionSelection` instance.

    Parameters
    ----------
    input_ : NengoObject
        Nengo object that retrieves the module input.
    type_ : nengo_spa.types.Type
        Type of the input.
    """

    def __init__(self, input_, type_):
        self.input = input_
        self.type = type_

    def __rrshift__(self, other):
        if not isinstance(other, Node):
            return NotImplemented
        if ActionSelection.active is None:
            infer_types(self, other)
            other.connect_to(self.input)
        else:
            return RoutedConnection(other, self)


class RoutedConnection(object):
    """Represents a routed connection from an AST node to a `.ModuleInput`.

    A routed connection is passed through an inhibitable channel to allow the
    routed connection to be disabled.

    Parameters
    ----------
    source : nengo_spa.ast.base.Node
        The AST node providing the source signal.
    sink : ModuleInput
        The module input that the source signal should be routed to.
    """

    _free_floating = set()

    def __init__(self, source, sink):
        self.type = infer_types(source, sink)
        self.source = source
        self.sink = sink
        RoutedConnection._free_floating.add(self)

    def connect_to(self, sink):
        return self.source.connect_to(sink)

    def construct(self):
        return self.connect_to(self.sink.input)

    @property
    def fixed(self):
        """Whether the source provides a fixed value."""
        return isinstance(self.source, Fixed)

    def transform(self):
        """For a fixed source, returns the transform to implement the output.
        """
        assert self.fixed
        if self.type == TScalar:
            return self.source.evaluate()
        else:
            return np.atleast_2d(self.source.evaluate().v).T


class ActionSelection(Mapping):
    """Implements an action selection system with basal ganglia and thalamus.

    The *ActionSelection* instance has to be used as context manager and each
    potential action is defined by an `.ifmax` call providing an expression
    for the utility value and any number of effects (routing of information)
    to activate when this utility value is highest of all.

    Attributes
    ----------
    active : ActionSelection
        Class attribute providing the currently active ActionSelection
        instance (if any).
    built : bool
        Indicates whether the action selection system has been built
        successfully.
    bg : nengo.Network
        Basal ganglia network. Available after the action selection system has
        been built.
    thalamus : nengo.Network
        Thalamus network. Available after the action selection system has
        been built.

    See Also
    --------

    nengo_spa.modules.BasalGanglia : Default basal ganglia network
    nengo_spa.modules.Thalamus : Default thalamus network

    Examples
    --------

    .. code-block:: python

        with ActionSelection():
            ifmax(dot(state, sym.A), sym.B >> state)
            ifmax(dot(state, sym.B), sym.C >> state)
            ifmax(dot(state, sym.C), sym.A >> state)

    This will route the *B* Semantic Pointer to *state* when *state* is more
    similar to *A* than any of the other Semantic Pointers. Similarly, *C*
    will be routed to *state* when *state* is *B*. Once, *state* is *C*, it
    will be reset to *A* and the cycle begins anew.

    Further action selection examples:

      * :ref:`/examples/question_control.ipynb`
      * :ref:`/examples/spa_parser.ipynb`
      * :ref:`/examples/spa_sequence.ipynb`
      * :ref:`/examples/spa_sequence_routed.ipynb`
    """

    active = None

    def __init__(self):
        super(ActionSelection, self).__init__()
        self.built = False
        self.bias = None
        self.bg = None
        self.thalamus = None
        self._utilities = []
        self._actions = []
        # Maps labels of actions to the index of that action
        self._name2idx = OrderedDict()

    def __enter__(self):
        assert not self.built
        if ActionSelection.active is None:
            ActionSelection.active = self
        else:
            raise SpaActionSelectionError(
                "Must not nest action selection contexts.")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        ActionSelection.active = None
        if exc_type is not None:
            return
        self._build()

    def _build(self):
        try:
            if len(RoutedConnection._free_floating) > 0:
                raise SpaActionSelectionError(
                    "All actions in an action selection context must be part "
                    "of an ifmax call.")
        finally:
            RoutedConnection._free_floating.clear()

        if len(self._utilities) <= 0:
            return

        self.bias = nengo.Node(1., label="bias")
        self.bg = dynamic.BasalGangliaRealization(len(self._utilities))
        self.thalamus = dynamic.ThalamusRealization(len(self._utilities))
        self.thalamus.connect_bg(self.bg)

        for index, utility in enumerate(self._utilities):
            self.bg.connect_input(utility, index=index)

        for index, action in enumerate(self._actions):
            for effect in action:
                if effect.fixed:
                    self.thalamus.connect_fixed(
                        index, effect.sink.input, transform=effect.transform())
                else:
                    self.thalamus.construct_gate(index, self.bias)
                    channel = self.thalamus.construct_channel(
                        effect.sink.input, effect.type)
                    effect.connect_to(channel.input)
                    self.thalamus.connect_gate(index, channel)
        self.built = True

    def __getitem__(self, key):
        if is_string(key):
            key = self._name2idx[key]
        return self._utilities[key]

    def __iter__(self):
        # Given not all actions have names, there will actions whose keys
        # will be numbers and not names.
        for i, (name, v) in enumerate(self._name2idx.items()):
            while i < v:
                yield i
                i += 1
            yield name
        for i in range(i + 1, len(self)):
            yield i

    def __len__(self):
        return len(self._actions)

    def add_action(self, name, *actions):
        assert ActionSelection.active is self
        utility = nengo.Node(size_in=1)
        if name is not None:
            self._name2idx[name] = len(self._actions)
        self._utilities.append(utility)
        self._actions.append(actions)
        RoutedConnection._free_floating.difference_update(actions)
        return utility


def ifmax(name, condition, *actions):
    """Defines a potential action within an `ActionSelection` context.

    Parameters
    ----------
    name : str
        Name for the action
    condition : nengo_spa.ast.base.Node
        The utility value for the given actions.
    actions : sequence of `RoutedConnection`
        The actions to activate if the given utility is the highest.

    Returns
    -------
    NengoObject
        Nengo object that can be connected to, to provide additional input to
        the utility value.
    """
    if ActionSelection.active is None:
        raise SpaActionSelectionError(
            "ifmax must be used within the context of an ActionSelection "
            "instance.")
    if condition.type != TScalar:
        raise SpaTypeError(
            "ifmax condition must evaluate to a scalar, but got {}.".format(
                condition.type))
    if any(not isinstance(a, RoutedConnection) for a in actions):
        raise SpaActionSelectionError(
            "ifmax actions must be routing expressions like 'a >> b'.")

    utility = ActionSelection.active.add_action(name, *actions)
    condition.connect_to(utility)
    return utility



import nengo
from nengo.config import Default
from nengo.exceptions import ValidationError
from nengo.params import IntParam, Parameter
from nengo.utils.compat import is_string
from nengo.utils.stdlib import checked_call
import numpy as np

from nengo_spa.ast.symbolic import PointerSymbol
from nengo_spa.network import Network
from nengo_spa.pointer import SemanticPointer
from nengo_spa.vocab import VocabularyOrDimParam


class SpArrayExtractor(object):
    def __init__(self, vocab):
        self.vocab = vocab

    def __call__(self, value):
        if isinstance(value, PointerSymbol):
            value = value.expr
        if is_string(value):
            value = self.vocab.parse(value)
        if isinstance(value, SemanticPointer):
            value = value.v
        return value


def make_sp_func(fn, vocab):
    def sp_func(t, v):
        return fn(t, SemanticPointer(v, vocab=vocab))
    return sp_func


def make_parse_func(fn, vocab):
    """Create a function that calls func and parses the output in vocab."""

    extractor = SpArrayExtractor(vocab)

    def parse_func(*args):
        return extractor(fn(*args))

    return parse_func


class TranscodeFunctionParam(Parameter):
    def coerce(self, obj, fn):
        fn = super(TranscodeFunctionParam, self).coerce(obj, fn)

        pointer_cls = (SemanticPointer, PointerSymbol)

        if fn is None:
            return fn
        elif callable(fn):
            return self.coerce_callable(obj, fn)
        elif not obj.input_vocab and (is_string(fn) or
                                      isinstance(fn, pointer_cls)):
            return fn
        else:
            raise ValidationError("Invalid output type {!r}".format(
                type(fn)), attr=self.name, obj=obj)

    def coerce_callable(self, obj, fn):
        t = 0.
        if obj.input_vocab is not None:
            args = (t, SemanticPointer(
                obj.input_vocab.dimensions, vocab=obj.input_vocab))
        elif obj.size_in is not None:
            args = (t, np.zeros(obj.size_in))
        else:
            args = (t,)

        _, invoked = checked_call(fn, *args)
        fn(*args)
        if not invoked:
            if obj.input_vocab is not None:
                raise ValidationError(
                    "Transcode function %r is expected to accept exactly 2 "
                    "arguments: time as a float, and a SemanticPointer",
                    attr=self.name, obj=obj)
            else:
                raise ValidationError(
                    "Transcode function %r is expected to accept exactly 1 or "
                    "2 arguments: time as a float, and optionally the input "
                    "data as NumPy array.", attr=self.name, obj=obj)
        return fn

    @classmethod
    def to_node_output(cls, fn, input_vocab=None, output_vocab=None):
        if fn is None:
            return None
        elif callable(fn):
            if input_vocab is not None:
                fn = make_sp_func(fn, input_vocab)
            if output_vocab is not None:
                fn = make_parse_func(fn, output_vocab)
            return fn
        elif is_string(fn) or isinstance(fn, (SemanticPointer, PointerSymbol)):
            return SpArrayExtractor(output_vocab)(fn)
        else:
            raise ValueError("Invalid output type {!r}".format(type(fn)))


class Transcode(Network):
    """Transcode from, to, and between Semantic Pointers.

    This can thought of the equivalent of a `nengo.Node` for Semantic Pointers.

    Either the *input_vocab* or the *output_vocab* argument must not be *None*.
    (If you want both arguments to be *None*, use a normal `nengo.Node`.)
    Which one of the parameters in the pairs *input_vocab/size_in* and
    *output_vocab/size_out* is not set to *None*, determines whether a Semantic
    Pointer input/output or a normal vector input/output is expected.

    Parameters
    ----------
    function : func, optional (Default: None)
        Function that transforms the input Semantic Pointer to an output
        Semantic Pointer. The function signature depends on *input_vocab*:

        * If *input_vocab* is *None*, the allowed signatures are the same as
          for a `nengo.Node`. Either ``function(t)`` or ``function(t, x)``
          where *t* (float) is the current simulation time and *x* (NumPy
          array) is the current input to transcode with size *size_in*.
        * If *input_vocab* is not *None*, the signature has to be
          ``function(t, sp)`` where *t* (float) is the current simulation time
          and *sp* (`.SemanticPointer`) is the current Semantic Pointer input.
          The associated vocabulary can be obtained via ``sp.vocab``.

        The allowed function return value depends on *output_vocab*:

        * If *output_vocab* is *None*, the return value must be a NumPy array
          (or equivalent) of size *size_out* or *None* (i.e. no return value)
          if *size_out* is *None*.
        * If *output_vocab* is not *None*, the return value can be either of:
          NumPy array, `.SemanticPointer` instance, or an SemanticPointer
          expression or symbolic expression as string that gets parsed with
          the *output_vocab*.
    input_vocab : Vocabulary, optional (Default: None)
        Input vocabulary. Mutually exclusive with *size_in*.
    output_vocab : Vocabulary, optional (Default: None)
        Output vocabulary. Mutually exclusive with *size_out*.
    size_in : int, optional (Default: None)
        Input size. Mutually exclusive with *input_vocab*.
    size_out : int, optional (Default: None)
        Output size. Mutually exclusive with *output_vocab*.
    kwargs : dict
        Additional keyword arguments passed to `nengo_spa.Network`.

    Attributes
    ----------
    input : nengo.Node
        Input.
    output : nengo.Node
        Output.
    """

    function = TranscodeFunctionParam(
        'function', optional=True, default=None, readonly=True)
    input_vocab = VocabularyOrDimParam(
        'input_vocab', optional=True, default=None, readonly=True)
    output_vocab = VocabularyOrDimParam(
        'output_vocab', optional=True, default=None, readonly=True)
    size_in = IntParam(
        'size_in', optional=True, default=None, readonly=True)
    size_out = IntParam(
        'size_out', optional=True, default=None, readonly=True)

    def __init__(
            self, function=Default, input_vocab=Default, output_vocab=Default,
            size_in=Default, size_out=Default, **kwargs):
        kwargs.setdefault('label', "Transcode")
        super(Transcode, self).__init__(**kwargs)

        # Vocabs need to be set before function which accesses vocab for
        # validation.
        self.input_vocab = input_vocab
        self.output_vocab = output_vocab
        self.size_in = size_in
        self.size_out = size_out

        if self.input_vocab is None and self.output_vocab is None:
            raise ValidationError(
                "At least one of input_vocab and output_vocab needs to be "
                "set. If neither the input nor the output is a Semantic "
                "Pointer, use a basic nengo.Node instead.", self)
        if self.input_vocab is not None and self.size_in is not None:
            raise ValidationError(
                "The input_vocab and size_in arguments are mutually "
                "exclusive.", 'size_in', self)
        if self.output_vocab is not None and self.size_out is not None:
            raise ValidationError(
                "The output_vocab and size_out arguments are mutually "
                "exclusive.", 'size_in', self)

        self.function = function

        node_size_in = (self.input_vocab.dimensions
                        if self.input_vocab is not None else self.size_in)
        node_size_out = (self.output_vocab.dimensions
                         if self.output_vocab is not None else self.size_out)
        if self.function is None:
            if node_size_in is None:
                node_size_in = self.output_vocab.dimensions
            node_size_out = None

        with self:
            self.node = nengo.Node(
                TranscodeFunctionParam.to_node_output(
                    self.function, self.input_vocab, self.output_vocab),
                size_in=node_size_in, size_out=node_size_out)
            self.input = self.node
            self.output = self.node

        if self.input_vocab is not None:
            self.declare_input(self.input, self.input_vocab)
        if self.output_vocab is not None:
            self.declare_output(self.output, self.output_vocab)



import pytest

from nengo_spa.exceptions import SpaTypeError
from nengo_spa.types import (
    coerce_types, TAnyVocab, TScalar, TAnyVocabOfDim, TVocabulary, Type)
from nengo_spa.vocab import Vocabulary


def test_coercion():
    v1 = TVocabulary(Vocabulary(16))

    assert coerce_types(TAnyVocab, TAnyVocab) is TAnyVocab
    assert coerce_types(TAnyVocab, TScalar) is TAnyVocab
    assert coerce_types(TAnyVocab, v1) == v1
    assert coerce_types(TScalar, TScalar) == TScalar
    assert coerce_types(TScalar, TScalar, TScalar) == TScalar
    assert coerce_types(TScalar, v1) == v1
    assert coerce_types(v1, v1) == v1
    assert coerce_types(TAnyVocab, v1, TScalar) == v1
    assert coerce_types(TScalar, TScalar, v1, TScalar, v1) == v1


def test_coercion_errors():
    with pytest.raises(SpaTypeError) as err:
        coerce_types(TVocabulary(Vocabulary(16)), TVocabulary(Vocabulary(16)))
    assert "Different vocabularies" in str(err.value)

    with pytest.raises(SpaTypeError) as err:
        coerce_types(TAnyVocabOfDim(16), TAnyVocabOfDim(32))
    assert "Dimensionality mismatch" in str(err.value)

    with pytest.raises(SpaTypeError) as err:
        coerce_types(Type('x'), Type('y'))
    assert "Incompatible types" in str(err.value)


def test_non_equality():
    v = Vocabulary(16)
    tv1 = TVocabulary(v)
    tv2 = TVocabulary(v)
    tvx = TVocabulary(Vocabulary(16))
    assert tv1 == tv2
    assert not (tv1 != tv2)
    assert tv1 != tvx
    assert not (tv1 == tvx)



import numpy as np
import pytest

import nengo
import nengo_spa as spa


def test_connect(Simulator, seed):
    with spa.Network(seed=seed) as model:
        model.buffer1 = spa.State(vocab=16)
        model.buffer2 = spa.State(vocab=16)
        model.buffer3 = spa.State(vocab=16)

        spa.sym.A >> model.buffer1
        model.buffer1 >> model.buffer2
        ~model.buffer1 >> model.buffer3

    with model:
        p2 = nengo.Probe(model.buffer2.output, synapse=0.03)
        p3 = nengo.Probe(model.buffer3.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.2)

    match = np.dot(sim.data[p2], model.buffer2.vocab.parse('A').v)
    assert match[199] > 0.8
    match = np.dot(sim.data[p3], model.buffer3.vocab.parse('~A').v)
    assert match[199] > 0.8


def test_transform(Simulator, seed):
    with spa.Network(seed=seed) as model:
        model.buffer1 = spa.State(vocab=32)
        model.buffer2 = spa.State(vocab=32)

        spa.sym.A >> model.buffer1
        model.buffer1 * spa.sym.B >> model.buffer2

    with model:
        p = nengo.Probe(model.buffer2.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.2)

    match = np.dot(sim.data[p], model.buffer2.vocab.parse('A*B').v)
    assert match[199] > 0.7


# FIXME test different populate arguments
def test_translate(Simulator, seed):
    with spa.Network(seed=seed) as model:
        model.buffer1 = spa.State(vocab=16)
        model.buffer2 = spa.State(vocab=32)

        spa.sym.A >> model.buffer1
        spa.translate(
            model.buffer1, model.buffer2.vocab, populate=True) >> model.buffer2

    with model:
        p = nengo.Probe(model.buffer2.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.2)

    match = np.dot(sim.data[p], model.buffer2.vocab.parse('A').v)
    assert match[199] > 0.8


def test_errors():
    # buffer2 does not exist
    with pytest.raises(AttributeError):
        with spa.Network() as model:
            model.buffer1 = spa.State(vocab=16)
            model.buffer1 >> model.buffer2


def test_direct(Simulator, seed):
    with spa.Network(seed=seed) as model:
        model.buffer1 = spa.State(vocab=16)
        model.buffer1.vocab.populate('A; B; C')
        model.buffer2 = spa.State(vocab=32)
        model.buffer2.vocab.populate('A; B; C')

        spa.sym.A >> model.buffer1
        spa.sym.B >> model.buffer2
        spa.sym.C >> model.buffer1
        spa.sym.C >> model.buffer2

    with model:
        p1 = nengo.Probe(model.buffer1.output, synapse=0.03)
        p2 = nengo.Probe(model.buffer2.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.2)

    match1 = np.dot(sim.data[p1], model.buffer1.vocab.parse('A+C').v)
    match2 = np.dot(sim.data[p2], model.buffer2.vocab.parse('B+C').v)
    # both values should be near 1.0 since buffer1 is driven to both A and C
    # and buffer2 is driven to both B and C.
    assert match1[199] > 0.75
    assert match2[199] > 0.75


def test_convolution(Simulator, plt, seed):
    D = 5
    with spa.Network(seed=seed) as model:
        model.config[spa.State].vocab = D
        model.config[spa.State].subdimensions = D
        model.inA = spa.State()
        model.inB = spa.State()
        model.outAB = spa.State()
        model.outABinv = spa.State()
        model.outAinvB = spa.State()
        model.outAinvBinv = spa.State()

        model.inA * model.inB >> model.outAB
        model.inA * ~model.inB >> model.outABinv
        ~model.inA * model.inB >> model.outAinvB
        ~model.inA * ~model.inB >> model.outAinvBinv
        nengo.Connection(nengo.Node([0, 1, 0, 0, 0]), model.inA.input)
        nengo.Connection(nengo.Node([0, 0, 1, 0, 0]), model.inB.input)

        pAB = nengo.Probe(model.outAB.output, synapse=0.03)
        pABinv = nengo.Probe(model.outABinv.output, synapse=0.03)
        pAinvB = nengo.Probe(model.outAinvB.output, synapse=0.03)
        pAinvBinv = nengo.Probe(model.outAinvBinv.output, synapse=0.03)

    for state in [
            model.inA,
            model.inB,
            model.outAB,
            model.outABinv,
            model.outAinvB,
            model.outAinvBinv]:
        for e in state.all_ensembles:
            e.radius = 1.

    with Simulator(model) as sim:
        sim.run(0.2)

    t = sim.trange()
    plt.subplot(4, 1, 1)
    plt.ylabel('A*B')
    plt.axhline(0.85, c='k')
    plt.plot(t, sim.data[pAB])
    plt.subplot(4, 1, 2)
    plt.ylabel('A*~B')
    plt.axhline(0.85, c='k')
    plt.plot(t, sim.data[pABinv])
    plt.subplot(4, 1, 3)
    plt.ylabel('~A*B')
    plt.axhline(0.85, c='k')
    plt.plot(t, sim.data[pAinvB])
    plt.subplot(4, 1, 4)
    plt.ylabel('~A*~B')
    plt.axhline(0.85, c='k')
    plt.plot(t, sim.data[pAinvBinv])

    # Check results.  Since A is [0,1,0,0,0] and B is [0,0,1,0,0], this means:
    #    ~A = [0,0,0,0,1]
    #    ~B = [0,0,0,1,0]
    #   A*B = [0,0,0,1,0]
    #  A*~B = [0,0,0,0,1]
    #  ~A*B = [0,1,0,0,0]
    # ~A*~B = [0,0,1,0,0]
    # (Remember that X*[1,0,0,0,0]=X (identity transform) and X*[0,1,0,0,0]
    #  is X rotated to the right once)

    # Ideal answer: A*B = [0,0,0,1,0]
    assert np.allclose(np.mean(sim.data[pAB][-10:], axis=0),
                       np.array([0, 0, 0, 1, 0]), atol=0.25)

    # Ideal answer: A*~B = [0,0,0,0,1]
    assert np.allclose(np.mean(sim.data[pABinv][-10:], axis=0),
                       np.array([0, 0, 0, 0, 1]), atol=0.25)

    # Ideal answer: ~A*B = [0,1,0,0,0]
    assert np.allclose(np.mean(sim.data[pAinvB][-10:], axis=0),
                       np.array([0, 1, 0, 0, 0]), atol=0.25)

    # Ideal answer: ~A*~B = [0,0,1,0,0]
    assert np.allclose(np.mean(sim.data[pAinvBinv][-10:], axis=0),
                       np.array([0, 0, 1, 0, 0]), atol=0.25)



import warnings

import nengo
from nengo.networks.ensemblearray import EnsembleArray
from nengo.params import Default, IntParam, NumberParam
from nengo.synapses import Lowpass, SynapseParam
import numpy as np

from nengo_spa.network import Network


# connection weights from (Gurney, Prescott, & Redgrave, 2001)
class Weights(object):
    mm = 1
    mp = 1
    me = 1
    mg = 1
    ws = 1
    wt = 1
    wm = 1
    wg = 1
    wp = 0.9
    we = 0.3
    e = 0.2
    ep = -0.25
    ee = -0.2
    eg = -0.2
    le = 0.2
    lg = 0.2

    @classmethod
    def str_func(cls, x):
        if x < cls.e:
            return 0
        return cls.mm * (x - cls.e)

    @classmethod
    def stn_func(cls, x):
        if x < cls.ep:
            return 0
        return cls.mp * (x - cls.ep)

    @classmethod
    def gpe_func(cls, x):
        if x < cls.ee:
            return 0
        return cls.me * (x - cls.ee)

    @classmethod
    def gpi_func(cls, x):
        if x < cls.eg:
            return 0
        return cls.mg * (x - cls.eg)


class BasalGanglia(Network):
    """Winner take all network, typically used for action selection.

    The basal ganglia network outputs approximately 0 at the dimension with
    the largest value, and is negative elsewhere.

    While the basal ganglia is primarily defined by its winner-take-all
    function, it is also organized to match the organization of the human
    basal ganglia. It consists of five ensembles:

    * Striatal D1 dopamine-receptor neurons (*strD1*)
    * Striatal D2 dopamine-receptor neurons (*strD2*)
    * Subthalamic nucleus (*stn*)
    * Globus pallidus internus / substantia nigra reticulata (*gpi*)
    * Globus pallidus externus (*gpe*)

    Interconnections between these areas are also based on known
    neuroanatomical connections. See [1]_ for more details, and [2]_ for
    the original non-spiking basal ganglia model by
    Gurney, Prescott & Redgrave that this model is based on.

    .. note:: The default `nengo.solvers.Solver` for the basal ganglia is
              `nengo.solvers.NnlsL2nz`, which requires SciPy. If SciPy is not
              installed, the global default solver will be used instead.

    Parameters
    ----------
    action_count : int
        Number of actions.
    n_neuron_per_ensemble : int, optional
        Number of neurons in each ensemble in the network.
    output_weight : float, optional
        A scaling factor on the output of the basal ganglia
        (specifically on the connection out of the GPi).
    input_bias : float, optional
        An amount by which to bias all dimensions of the input node.
        Biasing the input node is important for ensuring that all input
        dimensions are positive and easily comparable.
    ampa_synapse : Synapse, optional
        Synapse for connections corresponding to biological connections
        to AMPA receptors (i.e., connections from STN to to GPi and GPe).
    gaba_synapse : Synapse, optional
        Synapse for connections corresponding to biological connections
        to GABA receptors (i.e., connections from StrD1 to GPi, StrD2 to GPe,
        and GPe to GPi and STN).
    kwargs
        Passed through the `nengo_spa.Network`.

    Attributes
    ----------
    bias_input : nengo.Node or None
        If *input_bias* is non-zero, this node will be created to bias
        all of the dimensions of the input signal.
    gpe : nengo.networks.EnsembleArray
        Globus pallidus externus ensembles.
    gpi : nengo.networks.EnsembleArray
        Globus pallidus internus ensembles.
    input : nengo.Node
        Accepts the input signal.
    output : nengo.Node
        Provides the output signal.
    stn : nengo.networks.EnsembleArray
        Subthalamic nucleus ensembles.
    strD1 : nengo.networks.EnsembleArray
        Striatal D1 ensembles.
    strD2 : nengo.networks.EnsembleArray
        Striatal D2 ensembles.

    References
    ----------
    .. [1] Stewart, T. C., Choo, X., & Eliasmith, C. (2010).
       Dynamic behaviour of a spiking model of action selection in the
       basal ganglia. In Proceedings of the 10th international conference on
       cognitive modeling (pp. 235-40).
    .. [2] Gurney, K., Prescott, T., & Redgrave, P. (2001).
       A computational model of action selection in the basal
       ganglia. Biological Cybernetics 84, 401-423.
    """

    input_synapse = SynapseParam('input_synapse', default=Lowpass(0.002))
    ampa_synapse = SynapseParam('ampa_synapse', default=Lowpass(0.002))
    gaba_synapse = SynapseParam('gaba_synapse', default=Lowpass(0.008))
    n_neurons_per_ensemble = IntParam(
        'n_neurons_per_ensemble', default=100, low=1, readonly=True)
    output_weight = NumberParam('output_weight', default=-3., readonly=True)
    input_bias = NumberParam('input_bias', default=0., readonly=True)

    def __init__(
            self, action_count, n_neurons_per_ensemble=Default,
            output_weight=Default, input_bias=Default, ampa_synapse=Default,
            gaba_synapse=Default, **kwargs):
        kwargs.setdefault('label', "Basal ganglia")
        super(BasalGanglia, self).__init__(**kwargs)

        self.action_count = action_count
        self.n_neurons_per_ensemble = n_neurons_per_ensemble
        self.output_weight = output_weight
        self.input_bias = input_bias
        self.ampa_synapse = ampa_synapse
        self.gaba_synapse = gaba_synapse

        self.input_connections = {}
        # Affects all ensembles / connections in the BG
        # unless overwritten with general_config
        config = nengo.Config(nengo.Ensemble, nengo.Connection)
        config[nengo.Ensemble].radius = 1.5
        config[nengo.Ensemble].encoders = nengo.dists.Choice([[1]])
        try:
            # Best, if we have SciPy
            config[nengo.Connection].solver = nengo.solvers.NnlsL2nz()
        except ImportError:
            warnings.warn("SciPy is not installed, so BasalGanglia will "
                          "use the default decoder solver. Installing "
                          "SciPy may improve BasalGanglia performance.")

        ea_params = {'n_neurons': self.n_neurons_per_ensemble,
                     'n_ensembles': self.action_count}

        with self, config:
            self.strD1 = EnsembleArray(
                label="Striatal D1 neurons",
                intercepts=nengo.dists.Uniform(Weights.e, 1), **ea_params)
            self.strD2 = EnsembleArray(
                label="Striatal D2 neurons",
                intercepts=nengo.dists.Uniform(Weights.e, 1), **ea_params)
            self.stn = EnsembleArray(
                label="Subthalamic nucleus",
                intercepts=nengo.dists.Uniform(Weights.ep, 1), **ea_params)
            self.gpi = EnsembleArray(
                label="Globus pallidus internus",
                intercepts=nengo.dists.Uniform(Weights.eg, 1), **ea_params)
            self.gpe = EnsembleArray(
                label="Globus pallidus externus",
                intercepts=nengo.dists.Uniform(Weights.ee, 1), **ea_params)

            self.input = nengo.Node(label="input", size_in=self.action_count)
            self.output = nengo.Node(label="output", size_in=self.action_count)

            # add bias input (BG performs best in the range 0.5--1.5)
            if abs(self.input_bias) > 0.0:
                self.bias_input = nengo.Node(
                    np.ones(self.action_count) * self.input_bias,
                    label="basal ganglia bias")
                nengo.Connection(self.bias_input, self.input)

            # spread the input to StrD1, StrD2, and STN
            nengo.Connection(self.input, self.strD1.input, synapse=None,
                             transform=Weights.ws * (1 + Weights.lg))
            nengo.Connection(self.input, self.strD2.input, synapse=None,
                             transform=Weights.ws * (1 - Weights.le))
            nengo.Connection(self.input, self.stn.input, synapse=None,
                             transform=Weights.wt)

            # connect the striatum to the GPi and GPe (inhibitory)
            strD1_output = self.strD1.add_output('func_str', Weights.str_func)
            strD2_output = self.strD2.add_output('func_str', Weights.str_func)
            self.gaba = nengo.Network("GABAergic connections")
            self.gaba.config[nengo.Connection].synapse = self.gaba_synapse
            with self.gaba:
                nengo.Connection(strD1_output, self.gpi.input,
                                 transform=-Weights.wm)
                nengo.Connection(strD2_output, self.gpe.input,
                                 transform=-Weights.wm)

            # connect the STN to GPi and GPe (broad and excitatory)
            tr = Weights.wp * np.ones((self.action_count, self.action_count))
            stn_output = self.stn.add_output('func_stn', Weights.stn_func)
            self.ampa = nengo.Network("AMPAergic connectiions")
            self.ampa.config[nengo.Connection].synapse = self.ampa_synapse
            with self.ampa:
                nengo.Connection(stn_output, self.gpi.input, transform=tr)
                nengo.Connection(stn_output, self.gpe.input, transform=tr)

            # connect the GPe to GPi and STN (inhibitory)
            gpe_output = self.gpe.add_output('func_gpe', Weights.gpe_func)
            with self.gaba:
                nengo.Connection(
                    gpe_output, self.gpi.input, transform=-Weights.we)
                nengo.Connection(
                    gpe_output, self.stn.input, transform=-Weights.wg)

            # connect GPi to output (inhibitory)
            gpi_output = self.gpi.add_output('func_gpi', Weights.gpi_func)
            nengo.Connection(gpi_output, self.output, synapse=None,
                             transform=self.output_weight)

    def connect_input(self, source, transform=Default, index=None):
        self.input_connections[index] = nengo.Connection(
            source, self.input[index], transform=transform,
            synapse=self.input_synapse)



class SpaException(Exception):
    """A exception within the SPA subsystem."""


class SpaActionSelectionError(SpaException):
    """An error in the usage of the SPA action selection system."""


class SpaParseError(SpaException, ValueError):
    """An error encountered while parsing a SPA expression."""


class SpaTypeError(SpaException, ValueError):
    """The evaluation of types in an SPA expression was invalid."""



from collections import deque


class Peekable(object):
    """Wrapper for generator to allow peeking at coming elements.

    Parameters
    ----------
    generator : generator
        Generator to wrap.
    maxlen : int, optional
        Maximum number of items that can be looked ahead. If *None*, unlimited
        look-ahead is allowed.

    Attributes
    ----------
    generator : generator
        The wrapped generator.
    """

    def __init__(self, generator, maxlen=None):
        self.generator = generator
        self._buffer = deque(maxlen=maxlen)

    def __iter__(self):
        return self

    def __next__(self):
        if len(self._buffer) > 0:
            return self._buffer.popleft()
        else:
            return next(self.generator)

    def next(self):
        return self.__next__()

    def peek(self, n=1):
        """Looks *n* items ahead.

        Returns
        -------
        list
            List of coming items.
        """
        assert self._buffer.maxlen is None or n <= self._buffer.maxlen
        try:
            while len(self._buffer) < n:
                self._buffer.append(next(self.generator))
        except StopIteration:
            pass
        return [self._buffer[i] for i in range(min(n, len(self._buffer)))]



import warnings

import nengo
from nengo.exceptions import ValidationError
from nengo.utils.compat import is_iterable
from nengo.utils.network import with_self
import numpy as np


class IdentityEnsembleArray(nengo.Network):
    """An ensemble array optimized for representing the identity circular
    convolution vector besides random unit vectors.

    The ensemble array will use ensembles with *subdimensions* dimensions,
    except for the first *subdimensions* dimensions. These will be split into
    a one-dimensional ensemble for the first dimension and a *subdimensions-1*
    dimensional ensemble.

    Parameters
    ----------
    neurons_per_dimension : int
        Neurons per dimension.
    dimensions : int
        Total number of dimensions. Must be a multiple of *subdimensions*.
    subdimensions : int
        Maximum number of dimensions per ensemble.
    kwargs : dict
        Arguments to pass through to the `nengo.Network` constructor.

    Returns
    -------
    nengo.Network
        Network with attributes:

        * **input** (`nengo.Node`): Input node.
        * **output** (`nengo.Node`): Output node.
    """
    def __init__(
            self, neurons_per_dimension, dimensions, subdimensions, **kwargs):
        super(IdentityEnsembleArray, self).__init__(**kwargs)

        self.neurons_per_dimension = neurons_per_dimension
        self.dimensions = dimensions
        self.subdimensions = subdimensions

        self.neuron_input = None
        self.neuron_output = None

        cos_sim_dist = nengo.dists.CosineSimilarity(dimensions + 2)
        with self:
            self.input = nengo.Node(size_in=dimensions)
            self.output = nengo.Node(size_in=dimensions)

            self.first = nengo.Ensemble(neurons_per_dimension, 1)
            nengo.Connection(self.input[0], self.first, synapse=None)
            nengo.Connection(self.first, self.output[0], synapse=None)

            if subdimensions > 1:
                self.second = nengo.Ensemble(
                    neurons_per_dimension * (subdimensions - 1),
                    subdimensions - 1,
                    eval_points=cos_sim_dist, intercepts=cos_sim_dist)
                nengo.Connection(
                    self.input[1:subdimensions], self.second, synapse=None)
                nengo.Connection(
                    self.second, self.output[1:subdimensions], synapse=None)

            if dimensions > subdimensions:
                self.remainder = nengo.networks.EnsembleArray(
                    neurons_per_dimension * subdimensions,
                    dimensions // subdimensions - 1, subdimensions,
                    eval_points=cos_sim_dist, intercepts=cos_sim_dist)
                nengo.Connection(
                    self.input[subdimensions:], self.remainder.input,
                    synapse=None)
                nengo.Connection(
                    self.remainder.output, self.output[subdimensions:],
                    synapse=None)

    @with_self
    def add_neuron_input(self):
        """Adds a node providing input to the neurons of all ensembles.

        This node will be accessible through the *neuron_input* attribute.

        Returns
        -------
        nengo.Node
            The added node.
        """
        if self.neuron_input is not None:
            warnings.warn("neuron_input already exists. Returning.")
            return self.neuron_input

        if any(isinstance(e.neuron_type, nengo.Direct)
               for e in self.all_ensembles):
            raise ValidationError(
                "Ensembles use Direct neuron type. "
                "Cannot give neuron input to Direct neurons.",
                attr='all_ensembles[i].neuron_type', obj=self)

        self.neuron_input = nengo.Node(
            size_in=self.neurons_per_dimension * self.dimensions,
            label="neuron_input")

        i = 0
        for ens in self.all_ensembles:
            nengo.Connection(
                self.neuron_input[i:(i + ens.n_neurons)], ens.neurons,
                synapse=None)
            i += ens.n_neurons

        return self.neuron_input

    @with_self
    def add_neuron_output(self):
        """Adds a node providing neuron (non-decoded) output of all ensembles.

        This node will be accessible through the *neuron_output* attribute.

        Returns
        -------
        nengo.Node
            The added node.
        """
        if self.neuron_output is not None:
            warnings.warn("neuron_output already exists. Returning.")
            return self.neuron_output

        if any(isinstance(e.neuron_type, nengo.Direct)
               for e in self.all_ensembles):
            raise ValidationError(
                "Ensembles use Direct neuron type. "
                "Cannot get neuron output from Direct neurons.",
                attr='all_ensembles[i].neuron_type', obj=self)

        self.neuron_output = nengo.Node(
            size_in=self.neurons_per_dimension * self.dimensions,
            label="neuron_output")

        i = 0
        for ens in self.all_ensembles:
            nengo.Connection(
                ens.neurons, self.neuron_output[i:(i + ens.n_neurons)],
                synapse=None)
            i += ens.n_neurons

        return self.neuron_output

    @with_self
    def add_output(self, name, function, synapse=None, **conn_kwargs):
        """Adds a new decoded output.

        This will add the attribute named *name* to the object.

        Parameters
        ----------
        name : str
            Name of output. Must be a valid Python attribute name.
        function : func
            Function to decode.
        synapse : float or nengo.Lowpass
            Synapse to apply to the decoded connection to the returned output
            node.
        conn_kwargs : dict
            Additional keywword arguments to apply to the decoded connection.

        Returns
        -------
        nengo.Node
            Node providing the decoded output.
        """
        if is_iterable(function):
            function = list(function)
            if (len(function) != 3 and
                    len(function) != self.remainder.n_ensembles + 2):
                raise ValidationError(
                    "Must provide one function per ensemble or one function "
                    "each for the first ensemble, the second ensembles, and "
                    "all remaining ensembles.")
            first_fn = function[0]
            second_fn = function[1]
            remainder_fn = function[2:]
        else:
            first_fn = second_fn = remainder_fn = function

        first_size = np.asarray(first_fn(np.zeros(self.first.dimensions))).size
        second_size = np.asarray(
            second_fn(np.zeros(self.second.dimensions))).size
        remainder_start = first_size + second_size

        remainder_fn_out = self.remainder.add_output(name, remainder_fn)
        remainder_size = remainder_fn_out.size_out

        output = nengo.Node(
            size_in=first_size + second_size + remainder_size, label=name)
        setattr(self, name, output)

        nengo.Connection(
            self.first, output[:first_size], function=first_fn,
            synapse=synapse, **conn_kwargs)
        nengo.Connection(
            self.second, output[first_size:remainder_start],
            function=second_fn, synapse=synapse, **conn_kwargs)
        nengo.Connection(
            remainder_fn_out, output[remainder_start:], synapse=synapse,
            **conn_kwargs)

        return output



import sys

import nengo
import numpy as np
from numpy.testing import assert_allclose
import pytest

import nengo_spa as spa
from nengo_spa.ast.symbolic import PointerSymbol
from nengo_spa.exceptions import SpaTypeError
from nengo_spa.pointer import SemanticPointer
from nengo_spa.testing import sp_close


@pytest.mark.parametrize('op', ['-', '~'])
@pytest.mark.parametrize('suffix', ['', '.output'])
def test_unary_operation_on_module(Simulator, op, suffix, rng):
    vocab = spa.Vocabulary(64, rng=rng)
    vocab.populate('A')

    with spa.Network() as model:
        stimulus = spa.Transcode('A', output_vocab=vocab)  # noqa: F841
        x = eval(op + 'stimulus' + suffix)
        p = nengo.Probe(x.construct(), synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    assert sp_close(sim.trange(), sim.data[p], vocab.parse(op + 'A'), skip=0.3)


@pytest.mark.parametrize('op', ['+', '-', '*'])
@pytest.mark.parametrize('suffix', ['', '.output'])
def test_binary_operation_on_modules(Simulator, op, suffix, rng):
    vocab = spa.Vocabulary(64, rng=rng)
    vocab.populate('A; B')

    with spa.Network() as model:
        a = spa.Transcode('A', output_vocab=vocab)  # noqa: F841
        b = spa.Transcode('B', output_vocab=vocab)  # noqa: F841
        x = eval('a' + suffix + op + 'b' + suffix)
        p = nengo.Probe(x.construct(), synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    assert sp_close(
        sim.trange(), sim.data[p], vocab.parse('A' + op + 'B'), skip=0.3,
        atol=0.3)


@pytest.mark.parametrize('op', ['+', '-', '*'])
@pytest.mark.parametrize('order', ['AB', 'BA'])
def test_binary_operation_on_modules_with_pointer_symbol(
        Simulator, op, order, rng):
    vocab = spa.Vocabulary(64, rng=rng)
    vocab.populate('A; B')

    with spa.Network() as model:
        a = spa.Transcode('A', output_vocab=vocab)  # noqa: F841
        if order == 'AB':
            x = eval('a' + op + 'PointerSymbol("B")')
        elif order == 'BA':
            x = eval('PointerSymbol("B")' + op + 'a')
        else:
            raise ValueError('Invalid order argument.')
        p = nengo.Probe(x.construct(), synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    assert sp_close(
        sim.trange(), sim.data[p], vocab.parse(order[0] + op + order[1]),
        skip=0.3)


@pytest.mark.parametrize('op', ['+', '-', '*'])
@pytest.mark.parametrize('order', ['AB', 'BA'])
def test_binary_operation_on_modules_with_fixed_pointer(
        Simulator, op, order, rng):
    vocab = spa.Vocabulary(64, rng=rng)
    vocab.populate('A; B')
    b = SemanticPointer(vocab['B'].v)  # noqa: F841

    with spa.Network() as model:
        a = spa.Transcode('A', output_vocab=vocab)  # noqa: F841
        if order == 'AB':
            x = eval('a' + op + 'b')
        elif order == 'BA':
            x = eval('b' + op + 'a')
        else:
            raise ValueError('Invalid order argument.')
        p = nengo.Probe(x.construct(), synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    assert sp_close(
        sim.trange(), sim.data[p], vocab.parse(order[0] + op + order[1]),
        skip=0.3)


def test_complex_rule(Simulator, rng):
    vocab = spa.Vocabulary(64, rng=rng)
    vocab.populate('A; B; C; D')

    with spa.Network() as model:
        a = spa.Transcode('A', output_vocab=vocab)
        b = spa.Transcode('B', output_vocab=vocab)

        x = (0.5 * PointerSymbol('C') * a + 0.5 * PointerSymbol('D')) * (
            0.5 * b + a * 0.5)
        p = nengo.Probe(x.construct(), synapse=0.3)

    with nengo.Simulator(model) as sim:
        sim.run(0.5)

    assert sp_close(
        sim.trange(), sim.data[p],
        vocab.parse('(0.5 * C * A + 0.5 * D) * (0.5 * B + 0.5 * A)'),
        skip=0.3, normalized=True)


def test_transformed(Simulator, rng):
    vocab = spa.Vocabulary(64, rng=rng)
    vocab.populate('A; B')

    with spa.Network() as model:
        a = spa.Transcode('A', output_vocab=vocab)
        x = PointerSymbol('B') * a
        p = nengo.Probe(x.construct(), synapse=0.3)

    with nengo.Simulator(model) as sim:
        sim.run(0.5)

    assert sp_close(
        sim.trange(), sim.data[p], vocab.parse('B*A'), skip=0.3,
        normalized=True)


def test_transformed_and_pointer_symbol(Simulator, rng):
    vocab = spa.Vocabulary(64, rng=rng)
    vocab.populate('A; B')

    with spa.Network() as model:
        a = spa.Transcode('A', output_vocab=vocab)
        x = PointerSymbol('~B') * (PointerSymbol('B') * a)
        p = nengo.Probe(x.construct(), synapse=0.3)

    with nengo.Simulator(model) as sim:
        sim.run(0.5)

    assert sp_close(
        sim.trange(), sim.data[p], vocab.parse('~B * B * A'), skip=0.3,
        normalized=True)


def test_transformed_and_network(Simulator, rng):
    vocab = spa.Vocabulary(64, rng=rng)
    vocab.populate('A; B.unitary()')

    with spa.Network() as model:
        a = spa.Transcode('A', output_vocab=vocab)
        b = spa.Transcode('B', output_vocab=vocab)
        x = b * (PointerSymbol('~B') * a)
        p = nengo.Probe(x.construct(), synapse=0.3)

    with nengo.Simulator(model) as sim:
        sim.run(0.5)

    assert sp_close(
        sim.trange(), sim.data[p], vocab.parse('B * ~B * A'), skip=0.3,
        normalized=True)


def test_transformed_and_transformed(Simulator, rng):
    vocab = spa.Vocabulary(64, rng=rng)
    vocab.populate('A; B.unitary(); C')

    with spa.Network() as model:
        a = spa.Transcode('A', output_vocab=vocab)
        c = spa.Transcode('C', output_vocab=vocab)
        x = (PointerSymbol('B') * a) * (PointerSymbol('~B') * c)
        p = nengo.Probe(x.construct(), synapse=0.3)

    with nengo.Simulator(model) as sim:
        sim.run(0.5)

    assert sp_close(
        sim.trange(), sim.data[p], vocab.parse('B * A * ~B * C'), skip=0.3,
        normalized=True)


def test_pointer_symbol_with_dynamic_scalar(Simulator, rng):
    with spa.Network():
        scalar = spa.Scalar()
        with pytest.raises(SpaTypeError):
            PointerSymbol('A') * scalar


def test_dot(Simulator, rng):
    vocab = spa.Vocabulary(16, rng=rng)
    vocab.populate('A; B')

    with spa.Network() as model:
        a = spa.Transcode('A', output_vocab=vocab)
        b = spa.Transcode(
            lambda t: 'A' if t <= 0.5 else 'B', output_vocab=vocab)
        x = spa.dot(a, b)
        p = nengo.Probe(x.construct(), synapse=0.03)

    with nengo.Simulator(model) as sim:
        sim.run(1.)

    t = sim.trange()
    assert_allclose(sim.data[p][(0.3 < t) & (t <= 0.5)], 1., atol=.2)
    assert np.all(sim.data[p][0.8 < t] < 0.2)


def test_dot_with_fixed(Simulator, rng):
    vocab = spa.Vocabulary(16, rng=rng)
    vocab.populate('A; B')

    with spa.Network() as model:
        a = PointerSymbol('A')
        b = spa.Transcode(
            lambda t: 'A' if t <= 0.5 else 'B', output_vocab=vocab)
        x = spa.dot(a, b)
        p = nengo.Probe(x.construct(), synapse=0.03)

    with nengo.Simulator(model) as sim:
        sim.run(1.)

    t = sim.trange()
    assert_allclose(sim.data[p][(0.3 < t) & (t <= 0.5)], 1., atol=.2)
    assert np.all(sim.data[p][0.8 < t] < 0.2)


@pytest.mark.skipif(sys.version_info < (3, 5), reason="requires Python 3.5")
def test_dot_matmul(Simulator, rng):
    vocab = spa.Vocabulary(16, rng=rng)
    vocab.populate('A; B')

    with spa.Network() as model:
        a = spa.Transcode('A', output_vocab=vocab)  # noqa: F841
        b = spa.Transcode(  # noqa: F841
            lambda t: 'A' if t <= 0.5 else 'B', output_vocab=vocab)
        x = eval('a @ b')
        p = nengo.Probe(x.construct(), synapse=0.03)

    with nengo.Simulator(model) as sim:
        sim.run(1.)

    t = sim.trange()
    assert_allclose(sim.data[p][(0.3 < t) & (t <= 0.5)], 1., atol=.2)
    assert np.all(sim.data[p][0.8 < t] < 0.2)


@pytest.mark.skipif(sys.version_info < (3, 5), reason="requires Python 3.5")
def test_dot_with_fixed_matmul(Simulator, rng):
    vocab = spa.Vocabulary(16, rng=rng)
    vocab.populate('A; B')

    with spa.Network() as model:
        a = PointerSymbol('A')  # noqa: F841
        b = spa.Transcode(  # noqa: F841
            lambda t: 'A' if t <= 0.5 else 'B', output_vocab=vocab)
        x = eval('a @ b')
        p = nengo.Probe(x.construct(), synapse=0.03)

    with nengo.Simulator(model) as sim:
        sim.run(1.)

    t = sim.trange()
    assert_allclose(sim.data[p][(0.3 < t) & (t <= 0.5)], 1., atol=.2)
    assert np.all(sim.data[p][0.8 < t] < 0.2)


def test_dynamic_translate(Simulator, rng):
    v1 = spa.Vocabulary(64, rng=rng)
    v1.populate('A; B')
    v2 = spa.Vocabulary(64, rng=rng)
    v2.populate('A; B')

    with spa.Network() as model:
        source = spa.Transcode('A', output_vocab=v1)
        x = spa.translate(source, v2)
        p = nengo.Probe(x.construct(), synapse=0.03)

    with nengo.Simulator(model) as sim:
        sim.run(0.5)

    assert sp_close(sim.trange(), sim.data[p], v2['A'], skip=0.3, atol=0.2)



import numpy as np

import nengo
import nengo_spa as spa


def test_basic(Simulator, plt, seed):
    with nengo.Network(seed=seed) as model:
        bg = spa.BasalGanglia(action_count=5, seed=seed)
        input = nengo.Node([0.8, 0.4, 0.4, 0.4, 0.4], label="input")
        nengo.Connection(input, bg.input, synapse=None)
        p = nengo.Probe(bg.output, synapse=0.01)

    with Simulator(model) as sim:
        sim.run(0.2)

    t = sim.trange()
    output = np.mean(sim.data[p][t > 0.1], axis=0)

    plt.plot(t, sim.data[p])
    plt.ylabel("Output")

    assert output[0] > -0.1
    assert np.all(output[1:] < -0.8)


def test_basal_ganglia(Simulator, seed, plt):
    d = 64
    with spa.Network(seed=seed) as m:
        m.vision = spa.State(vocab=d)
        m.motor = spa.State(vocab=d)
        m.compare = spa.Compare(vocab=d)

        def input(t):
            if t < 0.1:
                return '0'
            elif t < 0.2:
                return 'CAT'
            elif t < 0.3:
                return 'DOG*~CAT'
            elif t < 0.4:
                return 'PARROT'
            elif t < 0.5:
                return 'MOUSE'
            else:
                return '0'
        m.encode = spa.Transcode(input, output_vocab=d)

        # test all acceptable condition formats
        with spa.ActionSelection() as actions:
            spa.ifmax(0.5, spa.sym.A >> m.motor)
            spa.ifmax(spa.dot(m.vision, spa.sym.CAT), spa.sym.B >> m.motor)
            spa.ifmax(
                spa.dot(m.vision * spa.sym.CAT, spa.sym.DOG),
                spa.sym.C >> m.motor)
            spa.ifmax(
                2 * spa.dot(m.vision, spa.sym.CAT * 0.5),
                spa.sym.D >> m.motor)
            spa.ifmax(spa.dot(m.vision, spa.sym.CAT) + 0.5 -
                      spa.dot(m.vision, spa.sym.CAT), spa.sym.E >> m.motor)
            spa.ifmax(
                spa.dot(m.vision, spa.sym.PARROT) + m.compare,
                spa.sym.F >> m.motor)
            spa.ifmax(
                0.5 * spa.dot(m.vision, spa.sym.MOUSE) + 0.5 * m.compare,
                spa.sym.G >> m.motor)
            spa.ifmax(
                (spa.dot(m.vision, spa.sym.MOUSE) - m.compare) * 0.5,
                spa.sym.H >> m.motor)

        m.encode >> m.vision
        spa.sym.SHOOP >> m.compare.input_a
        spa.sym.SHOOP >> m.compare.input_b
        bg = actions.bg

        p = nengo.Probe(bg.input, 'output', synapse=0.03)

    with Simulator(m) as sim:
        sim.run(0.5)
    t = sim.trange()

    plt.plot(t, sim.data[p])
    plt.legend(["A", "B", "C", "D", "E", "F", "G", "H"])
    plt.title('Basal Ganglia output')

    # assert the basal ganglia is prioritizing things correctly
    # Motor F
    assert sim.data[p][t == 0.4, 5] > 0.8
    # Motor G
    assert sim.data[p][t == 0.5, 6] > 0.8
    # Motor A
    assert 0.6 > sim.data[p][t == 0.1, 0] > 0.4
    # Motor B
    assert sim.data[p][t == 0.2, 1] > 0.8
    # Motor C
    assert sim.data[p][t == 0.3, 2] > 0.5

    # Motor B should be the same as Motor D
    assert np.allclose(sim.data[p][:, 1], sim.data[p][:, 3], atol=0.2)
    # Motor A should be the same as Motor E
    assert np.allclose(sim.data[p][:, 0], sim.data[p][:, 4], atol=0.2)


def test_scalar_product():
    with spa.Network() as model:
        model.scalar = spa.Scalar()
        with spa.ActionSelection():
            spa.ifmax(model.scalar * model.scalar, 1 >> model.scalar)
    # just testing network construction without exception here


def test_constructed_input_connections_are_accessible():
    with spa.Network() as model:
        model.config[spa.State].vocab = 16
        model.state1 = spa.State()
        model.state2 = spa.State()

        with spa.ActionSelection() as actions:
            spa.ifmax(
                spa.dot(model.state1, spa.sym.A), spa.sym.A >> model.state2)
        bg = actions.bg

        assert isinstance(bg.input_connections[0], nengo.Connection)



"""Support for unit testing SPA models."""

import numpy as np


def sp_close(
        t, data, target_sp, skip=0., duration=None, atol=0.2,
        normalized=False):
    """Test that the RMSE to a Semantic Pointer is below threshold.

    Parameters
    ----------
    t : (T,) array_like
        Time values for data in seconds, usually obtained with
        `nengo.Simulator.trange`.
    data : (T, D) array_like
        Simulation data for *T* timesteps and *D* dimensions.
    target_sp : SemanticPointer
        Target Semantic Pointer.
    skip : float, optional
        Amount of seconds to ignore at the beginning of the data.
    duration : float, optional
        Amount of seconds to consider after the skipped portion.
    atol : float, optional
        Absolute tolerated RMSE.
    normalize : bool, optional
        Whether to normalize the simulation data to unit length in each
        timestep.

    Returns
    -------
    bool
        *True* if *atol* is not exceeded during the considered time interval.
    """
    if duration is None:
        duration = np.max(t) - skip
    actual = data[(skip < t) & (t <= skip + duration)]
    expected = target_sp
    if normalized:
        actual /= np.expand_dims(np.linalg.norm(actual, axis=1), 1)
        expected = expected.normalized()
    return np.all(np.sqrt(np.sum(
        np.square(actual - expected.v), axis=1)) < atol)



import pytest

from nengo_spa.internal.generators import Peekable


def test_peekable():
    gen = Peekable(iter(range(5)), maxlen=2)
    assert gen.peek(n=2) == [0, 1]
    assert next(gen) == 0
    assert gen.peek() == [1]
    assert next(gen) == 1
    assert next(gen) == 2
    assert gen.peek() == [3]
    with pytest.raises(AssertionError):
        gen.peek(n=3)
    assert next(gen) == 3
    assert gen.peek(n=2) == [4]
    assert next(gen) == 4
    assert gen.peek() == []
    with pytest.raises(StopIteration):
        next(gen)


def test_peekable_no_maxlen():
    gen = Peekable(iter(range(5)), maxlen=None)
    assert gen.peek(n=20) == [0, 1, 2, 3, 4]
    for i in range(5):
        assert next(gen) == i
    with pytest.raises(StopIteration):
        next(gen)



import sys

import nengo
import numpy as np
from numpy.testing import assert_allclose, assert_equal
import pytest

import nengo_spa as spa
from nengo_spa import sym
from nengo_spa.ast.symbolic import FixedScalar, PointerSymbol
from nengo_spa.types import TVocabulary


def test_product_of_scalars(Simulator):
    with spa.Network() as model:
        stimulus = nengo.Node(0.5)
        a = spa.Scalar()
        b = spa.Scalar()
        nengo.Connection(stimulus, a.input)
        nengo.Connection(stimulus, b.input)
        x = a * b
        p = nengo.Probe(x.construct(), synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    assert_allclose(sim.data[p][sim.trange() > 0.3], .25, atol=0.2)


def test_unary_minus_on_scalar(rng):
    assert (-FixedScalar(1.)).evaluate() == -1.


def test_pointer_symbol_network_creation(rng):
    vocab = spa.Vocabulary(16, rng=rng)
    vocab.populate('A')

    with spa.Network():
        A = PointerSymbol('A', TVocabulary(vocab))
        node = A.construct()
    assert_equal(node.output, vocab['A'].v)


@pytest.mark.parametrize('op', ['-', '~'])
def test_unary_operation_on_pointer_symbol(op, rng):
    vocab = spa.Vocabulary(16, rng=rng)
    vocab.populate('A')

    with spa.Network():
        x = eval(op + "PointerSymbol('A', TVocabulary(vocab))")
        node = x.construct()
    assert_equal(node.output, vocab.parse(op + 'A').v)


@pytest.mark.parametrize('op', ['+', '-', '*'])
def test_binary_operation_on_pointer_symbols(op, rng):
    vocab = spa.Vocabulary(16, rng=rng)
    vocab.populate('A; B')

    with spa.Network():
        v = TVocabulary(vocab)  # noqa: F841
        x = eval("PointerSymbol('A', v)" + op + "PointerSymbol('B', v)")
        node = x.construct()
    assert_equal(node.output, vocab.parse('A' + op + 'B').v)


def test_pointer_symbol_mul_with_array():
    with pytest.raises(TypeError):
        PointerSymbol('X') * np.array([1, 2])


@pytest.mark.parametrize('op', ['+', '-'])
def test_additive_op_fixed_scalar_and_pointer_symbol(op, rng):
    vocab = spa.Vocabulary(16, rng=rng)
    vocab.populate('A')

    with spa.Network():
        with pytest.raises(TypeError):
            eval("2" + op + "PointerSymbol('A')")


def test_multiply_fixed_scalar_and_pointer_symbol(rng):
    vocab = spa.Vocabulary(16, rng=rng)
    vocab.populate('A')

    with spa.Network():
        x = 2 * PointerSymbol('A', TVocabulary(vocab))
        node = x.construct()
    assert_equal(node.output, vocab.parse('2 * A').v)


def test_fixed_dot(rng):
    vocab = spa.Vocabulary(16, rng=rng)
    vocab.populate('A; B')

    v = TVocabulary(vocab)
    assert_allclose(
        spa.dot(PointerSymbol('A', v), PointerSymbol('A', v)).evaluate(), 1.)
    assert spa.dot(
        PointerSymbol('A', v), PointerSymbol('B', v)).evaluate() <= 0.1


@pytest.mark.skipif(sys.version_info < (3, 5), reason="requires Python 3.5")
def test_fixed_dot_matmul(rng):
    vocab = spa.Vocabulary(16, rng=rng)
    vocab.populate('A; B')

    v = TVocabulary(vocab)  # noqa: F841
    assert_allclose(
        eval("PointerSymbol('A', v) @ PointerSymbol('A', v)").evaluate(), 1.)


def test_translate(rng):
    v1 = spa.Vocabulary(16, rng=rng)
    v1.populate('A; B')
    v2 = spa.Vocabulary(16, rng=rng)
    v2.populate('A; B')

    assert_allclose(
        spa.translate(PointerSymbol('A', TVocabulary(v1)), v2).evaluate().dot(
            v2['A']), 1., atol=0.2)


def test_reinterpret(rng):
    v1 = spa.Vocabulary(16, rng=rng)
    v1.populate('A; B')
    v2 = spa.Vocabulary(16, rng=rng)
    v2.populate('A; B')

    assert_equal(
        spa.reinterpret(PointerSymbol('A', TVocabulary(v1)), v2).evaluate().v,
        v1['A'].v)


def test_pointer_symbol_factory():
    ps = sym.A
    assert isinstance(ps, PointerSymbol)
    assert ps.expr == 'A'


def test_pointer_symbol_factory_expressions():
    ps = sym('A + B * C')
    assert isinstance(ps, PointerSymbol)
    assert ps.expr == (sym.A + sym.B * sym.C).expr



"""Basic classes for abstract syntax trees (ASTs) in Nengo SPA."""

from nengo.utils.compat import is_array

from nengo_spa.types import coerce_types, TAnyVocab, TVocabulary


def infer_types(*nodes):
    """Infers the most specific type for given nodes, sets end returns it.

    This determines the most specific type for given nodes. If it is a specific
    vocabulary, this vocabulary will be set for all less specific vocabulary
    types (but not scalar types) on the given node. Then the type will be
    returned.
    """
    type_ = coerce_types(*[n.type for n in nodes])
    if isinstance(type_, TVocabulary):
        for n in nodes:
            if TAnyVocab <= n.type < type_:
                n.type = type_
    return type_


class Node(object):
    """Base class for nodes in the AST for Nengo SPA operations.

    Parameters
    ----------
    type_ : nengo_spa.types.Type
        Type that the node evaluates to.
    """
    def __init__(self, type_):
        self.type = type_

    def connect_to(self, sink, **kwargs):
        """Implement the computation represented by the node and connect it.

        Parameters
        ----------
        sink : NengoObject
            Nengo object to connect to and transmit the result to.
        kwargs : dict
            Additional keyword arguments to pass to `nengo.Connection`.
        """
        raise NotImplementedError()

    def construct(self):
        """Implement the computation represented by the node.

        Returns
        -------
        NengoObject
            Usually the object providing the computation result as output, but
            can be something else in certain cases.
        """
        raise NotImplementedError()


class Noop(Node):
    """Node that has no effect."""

    def connect_to(self, sink, **kwargs):
        pass

    def construct(self):
        raise NotImplementedError("Noop nodes cannot be constructed.")


class Fixed(Node):
    """Base class for AST nodes that provide a fixed value."""

    def evaluate(self):
        """Return the fixed value that the node evaluates to."""
        raise NotImplementedError()


class TypeCheckedBinaryOp(object):
    """Decorator to check the type of the *other* parameter of an operator.

    If the *other* parameter is not an instance of *expected_type*,
    *NotImplemented* will be returned from the decorated method. If
    *conversion* is given it will be applied first.

    Parameters
    ----------
    expected_type : class
        Type for which the operator is implemented.
    conversion : function, optional
        Used to convert *other* before checking its type.
    """

    __slots__ = ['expected_type', 'conversion']

    def __init__(self, expected_type, conversion=None):
        self.expected_type = expected_type
        self.conversion = conversion

    def __call__(self, fn):
        def checked(inst, other):
            if self.conversion is not None:
                other = self.conversion(other)
            if is_array(other):
                raise TypeError("Bare array not allowed in SPA operation.")
            elif not isinstance(other, self.expected_type):
                return NotImplemented
            return fn(inst, other)
        return checked



from collections import Mapping
from keyword import iskeyword
import re
import warnings

import nengo
from nengo.exceptions import NengoWarning, ValidationError
from nengo.utils.compat import is_number, is_integer, range
import numpy as np

from nengo_spa import pointer
from nengo_spa.exceptions import SpaParseError
from nengo_spa.pointer import Identity


valid_sp_regex = re.compile('^[A-Z][_a-zA-Z0-9]*$')
reserved_sp_names = {'None', 'True', 'False'}


class Vocabulary(Mapping):
    """A collection of semantic pointers, each with their own text label.

    The Vocabulary can also act as a dictionary, with keys as the names
    of the semantic pointers and values as the `.SemanticPointer` objects
    themselves. The names of Semantic Pointers must be valid Python 2
    identifiers starting with a capital letter.

    Parameters
    -----------
    dimensions : int
        Number of dimensions for each semantic pointer.
    strict : bool, optional
        Whether to automatically create missing semantic pointers. If a
        non-strict vocabulary is asked for a pointer that does not exist within
        the vocabulary, the missing pointer will be automatically added to the
        vocabulary. A strict vocabulary will throw an error if asked for a
        pointer that does not exist in the vocabulary.
    max_similarity : float, optional
        When randomly generating pointers, ensure that the cosine of the
        angle between the new pointer and all existing pointers is less
        than this amount. If the system is unable to find such a pointer
        after 100 tries, a warning message is printed.
    rng : numpy.random.RandomState, optional
        The random number generator to use to create new vectors.
    name : str
        A name to display in the string representation of this vocabulary.

    Attributes
    ----------
    keys : sequence
        The names of all known semantic pointers (e.g., ``['A', 'B', 'C']``).
    max_similarity : float
        When randomly generating pointers, ensure that the cosine of the
        angle between the new pointer and all existing pointers is less
        than this amount. If the system is unable to find such a pointer
        after 100 tries, a warning message is printed.
    strict : bool
        Whether to automatically create missing semantic pointers. If a
        non-strict vocabulary is asked for a pointer that does not exist within
        the vocabulary, the missing pointer will be automatically added to the
        vocabulary. A strict vocabulary will throw an error if asked for a
        pointer that does not exist in the vocabulary.
    vectors : ndarray
        All of the semantic pointer vectors in a matrix, in the same order
        as in `keys`.
    """

    def __init__(
            self, dimensions, strict=True, max_similarity=0.1, rng=None,
            name=None):

        if not is_integer(dimensions) or dimensions < 1:
            raise ValidationError("dimensions must be a positive integer",
                                  attr='dimensions', obj=self)
        self.dimensions = dimensions
        self.strict = strict
        self.max_similarity = max_similarity
        self._key2idx = {}
        self._keys = []
        self._vectors = np.zeros((0, dimensions), dtype=float)
        self.rng = rng
        self.name = name

    @property
    def vectors(self):
        v = self._vectors.view()
        v.setflags(write=False)
        return v

    def __str__(self):
        name = '' if self.name is None else '"{}" '.format(self.name)
        return '{}-dimensional vocab {}at 0x{:x}'.format(
            self.dimensions, name, id(self))

    def create_pointer(self, attempts=100, transform=None):
        """Create a new semantic pointer and add it to the vocabulary.

        This will take into account the `max_similarity` attribute.  If a
        pointer satisfying max_similarity is not generated after the specified
        number of attempts, the candidate pointer with lowest maximum cosine
        similarity with all existing pointers is returned.

        Parameters
        ----------
        attempts : int, optional
            Maximum number of attempts to create a Semantic Pointer not
            exceeding `max_similarity`.
        transform : str, optional
            A transform to apply to the generated vector. Needs to be the name
            of a method of `.SemanticPointer`. Currently, the only sensible
            value is 'unitary'.

        Returns
        -------
        SemanticPointer
            The generated Semantic Pointer.
        """
        best_p = None
        best_sim = np.inf
        for _ in range(attempts):
            p = pointer.SemanticPointer(
                self.dimensions, rng=self.rng, vocab=self)
            if transform is not None:
                p = eval('p.' + transform, dict(self), {'p': p})
            if len(self) == 0:
                best_p = p
                break
            else:
                p_sim = np.max(np.dot(self._vectors, p.v))
                if p_sim < best_sim:
                    best_p = p
                    best_sim = p_sim
                    if p_sim < self.max_similarity:
                        break
        else:
            warnings.warn(
                'Could not create a semantic pointer with '
                'max_similarity=%1.2f (D=%d, M=%d, similarity=%1.2f)'
                % (self.max_similarity, self.dimensions,
                   len(self._key2idx), best_sim))
        return best_p

    def __contains__(self, key):
        return key in self._key2idx

    def __len__(self):
        return len(self._vectors)

    def __iter__(self):
        return iter(self._keys)

    def __getitem__(self, key):
        """Return the semantic pointer with the requested name."""
        # __tracebackhide__ is used in py.test to hide stack frames from the
        # traceback. That means py.test might try to look up this attribute
        # in a test which will result in an exception hiding the actual
        # exception. By raising a KeyError we indicate that there is no
        # __tracebackhide__ attribute on this object and preserve the relevant
        # exception.
        if key == '__tracebackhide__':
            raise KeyError()
        if not self.strict and key not in self:
            self.add(key, self.create_pointer())
        return pointer.SemanticPointer(
            self._vectors[self._key2idx[key]], vocab=self)

    def __hash__(self):
        return hash(id(self))

    def add(self, key, p):
        """Add the semantic pointer *p* to the vocabulary.

        Parameters
        ----------
        key : str
            Name of the Semantic Pointer. Must be a valid Python 2 identifier
            starting with a capital letter.
        p : SemanticPointer or array_like
            Semantic Pointer to add.
        """
        if (not valid_sp_regex.match(key) or iskeyword(key) or
                key in reserved_sp_names):
            raise SpaParseError(
                "Invalid Semantic Pointer name {!r}. Valid names are valid "
                "Python 2 identifiers beginning with a capital letter.".format(
                    key))
        if not isinstance(p, pointer.SemanticPointer):
            p = pointer.SemanticPointer(p, vocab=self)

        if key in self._key2idx:
            raise ValidationError("The semantic pointer %r already exists"
                                  % key, attr='', obj=self)
        if p.vocab is not None and p.vocab is not self:
            raise ValidationError(
                "Cannot add a semantic pointer that belongs to a different "
                "vocabulary.", attr='', obj=self)

        self._key2idx[key] = len(self._key2idx)
        self._keys.append(key)
        self._vectors = np.vstack([self._vectors, p.v])

    def populate(self, pointers):
        """Populate the vocabulary with semantic pointers given an expression.

        In its most basic form *pointers* is a string of names separated with
        ``;``::

            vocab.populate('A; B; C')

        Semantic Pointers can be constructed from other Semantic Pointers::

            vocab.populate('A; B; C = 0.3 * A + 1.4 * C')

        Those constructed Semantic Pointers are **not** normalized to
        unit-length. This can be done by appending a ``normalized()`` call.
        In the same way unitary Semantic Pointers can be obtained with
        ``unitary()``::

            vocab.populate('A.unitary(); B; C = (A+B).normalized()')

        Parameters
        ----------
        pointers : string
            The expression defining the semantic pointers to
            add to the vocabulary.
        """
        if len(pointers.strip()) <= 0:
            return  # Do nothing (and don't fail) for empty string.

        for p_expr in pointers.split(';'):
            assign_split = p_expr.split('=', 1)
            modifier_split = p_expr.split('.', 1)
            if len(assign_split) > 1:
                name, value_expr = assign_split
                value = eval(value_expr.strip(), {}, self)
            elif len(modifier_split) > 1:
                name = modifier_split[0]
                value = self.create_pointer(transform=modifier_split[1])
            else:
                name = p_expr
                value = self.create_pointer()
            self.add(name.strip(), value)

    def parse(self, text):
        """Evaluate a text string and return the corresponding SemanticPointer.

        This uses the Python ``eval()`` function, so any Python operators that
        have been defined for SemanticPointers are valid (``+``, ``-``, ``*``,
        ``~``, ``()``). Valid semantic pointer terms must start
        with a capital letter.

        If the expression returns a scalar (int or float), a scaled version
        of the identity SemanticPointer will be returned.
        """

        # The following line does everything.  Note that self is being
        # passed in as the locals dictionary, and thanks to the __getitem__
        # implementation, this will automatically create new semantic
        # pointers as needed.
        try:
            value = eval(text, {}, self)
        except NameError as err:
            raise SpaParseError(
                "Error parsing expression {expr!r} with {vocab}: {msg}".format(
                    expr=text, vocab=self, msg=str(err)))

        if is_number(value):
            value *= Identity(self.dimensions)
        elif not isinstance(value, pointer.SemanticPointer):
            raise SpaParseError(
                "The result of parsing '%s' is not a SemanticPointer." % text)
        return value

    def parse_n(self, *texts):
        """Applies `parse` to each item in *texts* and returns the result."""
        return [self.parse(t) for t in texts]

    def dot(self, v):
        """Returns the dot product with all terms in the Vocabulary.

        Parameters
        ----------
        v : SemanticPointer or array_like
            SemanticPointer to calculate dot product with.
        """
        if isinstance(v, pointer.SemanticPointer):
            v = v.v
        return np.dot(self._vectors, v)

    def transform_to(self, other, populate=None, keys=None, solver=None):
        """Create a linear transform from one Vocabulary to another.

        This is simply the sum of the outer products of the corresponding
        terms in each Vocabulary if no *solver* is given, otherwise a
        least-squares solution will be obtained.

        Parameters
        ----------
        other : Vocabulary
            The vocabulary to translate into.
        populate : Boolean
            Whether to add the missing keys from the original vocabulary
            to the new target vocabulary.
        keys : list, optional
            Limits the Semantic Pointers considered from the original
            vocabulary if given.
        solver: callable
            Solver to obtain least-squares solution to map one vocabulary to
            the other.
        """
        if keys is None:
            keys = self._keys
        keys = set(keys)

        missing_keys = set(k for k in keys if k not in other)

        if len(missing_keys) > 0:
            if populate is None:
                warnings.warn(NengoWarning(
                    "The transform_to source vocabulary has keys not existent "
                    "in the target vocabulary. These will be ignored. Use the "
                    "`populate=False` keyword argument to silence this "
                    "warning or `populate=True` to automatically add missing "
                    "keys to the target vocabulary."))
            elif populate:
                other.populate(';'.join(missing_keys))
                missing_keys = set()

        from_vocab = self.create_subset(keys - missing_keys).vectors
        to_vocab = other.create_subset(keys - missing_keys).vectors
        if solver is None:
            return np.dot(to_vocab.T, from_vocab)
        else:
            return solver(from_vocab, to_vocab)[0].T

    def create_subset(self, keys):
        """Returns a subset of this vocabulary.

        Creates and returns a subset of the current vocabulary that contains
        all the semantic pointers found in keys.

        Parameters
        ----------
        keys : sequence
            List or set of semantic pointer names to be copied over to the
            new vocabulary.
        """
        # Make new Vocabulary object
        subset = Vocabulary(self.dimensions, self.strict, self.max_similarity,
                            self.rng)

        # Copy over the new keys
        for key in keys:
            subset.add(key, self[key].reinterpret(subset))

        return subset


class VocabularyMap(Mapping):
    """Maps dimensionalities to corresponding vocabularies.

    Acts like a Python dictionary.

    Parameters
    ----------
    vocabs : sequence of Vocabulary
        A list of vocabularies to add to the mapping. The dimensionalities
        will be determined from the vocabulary objects.
    rng : numpy.random.RandomState
        Random number generator to use for newly created vocabularies (with
        `.get_or_create`).
    """

    def __init__(self, vocabs=None, rng=None):
        if vocabs is None:
            vocabs = []
        self.rng = rng

        self._vocabs = {}
        try:
            for vo in vocabs:
                self.add(vo)
        except (AttributeError, TypeError):
            raise ValueError(
                "The `vocabs` argument requires a list of Vocabulary "
                "instances or `None`.")

    def add(self, vocab):
        """Add a vocabulary to the map.

        The dimensionality will be determined from the vocabulary.

        Parameters
        ----------
        vocab : Vocaublary
            Vocabulary to add.
        """
        if vocab.dimensions in self._vocabs:
            warnings.warn("Duplicate vocabularies with dimension %d. "
                          "Using the last entry in the vocab list with "
                          "that dimensionality." % vocab.dimensions)
        self._vocabs[vocab.dimensions] = vocab

    def __delitem__(self, dimensions):
        del self._vocabs[dimensions]

    def discard(self, vocab):
        """Discard (remove) a vocabulary from the mapping.

        Parameters
        ----------
        vocab : int or Vocabulary
            If an integer is given, the vocabulary associated to the
            dimensionality will be discarded. If a `.Vocabulary` is given, that
            specific instance will be discarded.
        """
        if isinstance(vocab, int):
            del self._vocabs[vocab]
        elif self._vocabs.get(vocab.dimensions, None) is vocab:
            del self._vocabs[vocab.dimensions]

    def __getitem__(self, dimensions):
        return self._vocabs[dimensions]

    def get_or_create(self, dimensions):
        """Gets or creates a vocabulary of given dimensionality.

        If the mapping already maps the given dimensionality to a vocabulary,
        it will be returned. Otherwise, a new vocabulary will be created,
        added to the mapping, and returned.

        Parameters
        ----------
        dimensions : int
            Dimensionality of vocabulary to return.

        Returns
        -------
        Vocabulary
            Vocabulary of given dimensionality.
        """
        if dimensions not in self._vocabs:
            self._vocabs[dimensions] = Vocabulary(
                dimensions, strict=False, rng=self.rng)
        return self._vocabs[dimensions]

    def __iter__(self):
        return iter(self._vocabs)

    def __len__(self):
        return len(self._vocabs)

    def __contains__(self, vocab):
        if isinstance(vocab, int):
            return vocab in self._vocabs
        else:
            return (vocab.dimensions in self._vocabs and
                    self._vocabs[vocab.dimensions] is vocab)


class VocabularyMapParam(nengo.params.Parameter):
    """Nengo parameter that accepts `.VocabularyMap` instances.

    Sequences of `.Vocabulary` will be coerced to `.VocabularyMap`.
    """

    def coerce(self, instance, vocab_set):
        vocab_set = super(VocabularyMapParam, self).coerce(instance, vocab_set)

        if vocab_set is not None and not isinstance(vocab_set, VocabularyMap):
            try:
                vocab_set = VocabularyMap(vocab_set)
            except ValueError:
                raise ValidationError(
                    "Must be of type 'VocabularyMap' or compatible "
                    "(got type %r)."
                    % type(vocab_set).__name__, attr=self.name, obj=instance)

        return vocab_set


class VocabularyOrDimParam(nengo.params.Parameter):
    """Nengo parameter that accepts `.Vocabulary` or integer dimensionality.

    If an integer is assigned, the vocabulary will retrieved from the
    instance's *vocabs* attribute with *vocabs.get_or_create(dimensions)*.
    Thus, a class using *VocabularyOrDimParam* should also have an attribute
    *vocabs* of type `VocabularyMap`.
    """

    coerce_defaults = False

    def coerce(self, instance, value):
        value = super(VocabularyOrDimParam, self).coerce(instance, value)

        if value is not None:
            if is_integer(value):
                if value < 1:
                    raise ValidationError(
                        "Vocabulary dimensionality must be at least 1.",
                        attr=self.name, obj=instance)
                value = instance.vocabs.get_or_create(value)
            elif not isinstance(value, Vocabulary):
                raise ValidationError(
                    "Must be of type 'Vocabulary' or an integer (got type %r)."
                    % type(value).__name__, attr=self.name, obj=instance)
        return value



import nengo
from nengo.params import Default, IntParam

from nengo_spa.network import Network


class Product(Network):
    """Multiplies two scalars.

    Parameters
    ----------
    n_neurons : int, optional (Default: 200)
        Number of neurons to use in product computation.
    kwargs : dict
        Keyword arguments passed through to `nengo_spa.Network`.

    Attributes
    ----------
    input_a : nengo.Node
        First input.
    input_b : nengo.Node
        Second input.
    output : nengo.Node
        Output.
    """

    n_neurons = IntParam('n_neurons', default=200, low=1, readonly=True)

    def __init__(self, n_neurons=Default, **kwargs):
        kwargs.setdefault('label', "Product")
        super(Product, self).__init__(**kwargs)

        self.n_neurons = n_neurons

        with self:
            self.product = nengo.networks.Product(self.n_neurons, 1)

        self.input_a = self.product.input_a
        self.input_b = self.product.input_b
        self.output = self.product.output

        self.declare_input(self.input_a, None)
        self.declare_input(self.input_b, None)
        self.declare_output(self.output, None)



import nengo
import nengo_spa as spa


def test_basic():
    with spa.Network():
        compare = spa.Compare(vocab=16)

    vocab_a = spa.Network.get_input_vocab(compare.input_a)
    vocab_b = spa.Network.get_input_vocab(compare.input_b)
    # all inputs should share the same vocab
    assert vocab_a is vocab_b
    assert vocab_a.dimensions == 16
    # output should have no vocab
    assert spa.Network.get_output_vocab(compare.output) is None


def test_run(Simulator, seed):
    with spa.Network(seed=seed) as model:
        model.compare = spa.Compare(vocab=16)
        model.compare.vocab.populate('A; B')

        def inputA(t):
            if 0 <= t < 0.1:
                return 'A'
            else:
                return 'B'

        model.input = spa.Transcode(inputA, output_vocab=16)
        model.input >> model.compare.input_a
        spa.sym.A >> model.compare.input_b

    with model:
        p = nengo.Probe(model.compare.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.2)

    assert sim.data[p][100] > 0.8
    assert sim.data[p][199] < 0.2



import numpy as np
from numpy.testing import assert_allclose
import pytest

import nengo
import nengo_spa as spa
from nengo_spa.exceptions import SpaTypeError
from nengo_spa.examine import similarity
from nengo_spa.network import create_inhibit_node
from nengo_spa.vocab import VocabularyMap


class SpaCommunicationChannel(spa.Network):
    def __init__(
            self, dimensions, label=None, seed=None, add_to_container=None):
        super(SpaCommunicationChannel, self).__init__(
            label, seed, add_to_container)

        with self:
            self.state_in = spa.State(dimensions)
            self.state_out = spa.State(dimensions)
            self.secondary = spa.State(dimensions)

            self.state_in >> self.state_out

        self.input = self.state_in.input
        self.input_secondary = self.secondary.input
        self.output = self.state_out.output
        self.output_secondary = self.secondary.output


def test_spa_vocab():
    # create a model without a vocab and check that it is empty
    model = spa.Network()
    assert len(model.vocabs) == 0

    # create a model with a vocab and check that it's filled
    va = spa.Vocabulary(16)
    va.populate("PANTS")
    vb = spa.Vocabulary(32)
    vb.populate("SHOES")
    model = spa.Network(vocabs=VocabularyMap([va, vb]))
    assert list(model.vocabs[16].keys()) == ["PANTS"]
    assert list(model.vocabs[32].keys()) == ["SHOES"]

    # warning on vocabs with duplicate dimensions
    vc = spa.Vocabulary(16)
    vc.populate("SOCKS")
    with pytest.warns(UserWarning):
        model = spa.Network(vocabs=VocabularyMap([va, vb, vc]))
    assert list(model.vocabs[16].keys()) == ["SOCKS"]
    assert list(model.vocabs[32].keys()) == ["SHOES"]


def test_hierarchical(Simulator, seed, plt):
    d = 32

    with spa.Network(seed=seed) as model:
        model.comm_channel = SpaCommunicationChannel(d)
        model.out = spa.State(d)

        spa.sym.A >> model.comm_channel
        model.comm_channel >> model.out

        p = nengo.Probe(model.out.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    t = sim.trange() > 0.2
    v = model.vocabs[d].parse('A').v

    plt.plot(sim.trange(), similarity(sim.data[p], v))
    plt.xlabel("t [s]")
    plt.ylabel("Similarity")

    assert np.mean(similarity(sim.data[p][t], v)) > 0.8


def test_hierarchical_actions(Simulator, seed, plt):
    d = 32

    with spa.Network(seed=seed) as model:
        model.comm_channel = SpaCommunicationChannel(d)
        model.out = spa.State(d)

        spa.sym.A >> model.comm_channel.state_in
        model.comm_channel.state_out >> model.out

        p = nengo.Probe(model.out.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    t = sim.trange() > 0.2
    v = model.vocabs[d].parse('A').v

    plt.plot(sim.trange(), similarity(sim.data[p], v))
    plt.xlabel("t [s]")
    plt.ylabel("Similarity")

    assert np.mean(similarity(sim.data[p][t], v)) > 0.8


def test_vocab_config():
    with spa.Network() as model:
        with spa.Network() as model.shared_vocabs:
            pass
        with spa.Network(
                vocabs=spa.vocab.VocabularyMap()) as model.non_shared_vocabs:
            pass

    assert model.shared_vocabs.vocabs is model.vocabs
    assert model.non_shared_vocabs.vocabs is not model.vocabs


def test_no_magic_vocab_transform():
    d = 16
    v1 = spa.Vocabulary(d)
    v2 = spa.Vocabulary(d)

    with spa.Network() as model:
        model.a = spa.State(vocab=v1)
        model.b = spa.State(vocab=v2)
        with pytest.raises(SpaTypeError):
            model.a >> model.b


@pytest.mark.parametrize('d1,d2,method,lookup', [
    (16, 16, 'reinterpret(a, v2)', 'v1'),
    (16, 16, 'reinterpret(a)', 'v1'),
    (16, 32, 'translate(a, v2)', 'v2'),
    (16, 32, 'translate(a, v2, solver=nengo.solvers.Lstsq())', 'v2')])
def test_casting_vocabs(d1, d2, method, lookup, Simulator, plt, rng):
    v1 = spa.Vocabulary(d1, rng=rng)
    v1.populate('A')
    v2 = spa.Vocabulary(d2, rng=rng)
    v2.populate('A')

    with spa.Network() as model:
        a = spa.State(vocab=v1)
        b = spa.State(vocab=v2)
        spa.sym.A >> a
        eval("spa.%s" % method) >> b
        p = nengo.Probe(b.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    t = sim.trange() > 0.2
    v = locals()[lookup].parse('A').v

    plt.plot(sim.trange(), spa.similarity(sim.data[p], v))
    plt.xlabel("t [s]")
    plt.ylabel("Similarity")

    assert np.mean(spa.similarity(sim.data[p][t], v)) > 0.8


def test_copy_spa(RefSimulator):
    with spa.Network() as original:
        original.state = spa.State(16)
        spa.sym.A >> original.state

    cp = original.copy()

    # Check that it still builds.
    with RefSimulator(cp):
        pass


def test_create_inhibit_node(Simulator, plt):
    with nengo.Network() as model:
        ea = nengo.networks.EnsembleArray(10, 10)
        bias = nengo.Node(1)
        inhibit_node = create_inhibit_node(ea)
        nengo.Connection(bias, inhibit_node)
        nengo.Connection(
            bias, ea.input, transform=np.ones((ea.n_ensembles, 1)))
        p = nengo.Probe(ea.output, synapse=0.01)

    with Simulator(model) as sim:
        sim.run(0.3)

    plt.plot(sim.trange(), sim.data[p])
    plt.xlabel("Time [s]")

    assert_allclose(sim.data[p][sim.trange() > 0.1], 0., atol=1e-3)


def test_instance_config():
    with spa.Network() as net:
        ens = nengo.Ensemble(10, 1)
        net.config[nengo.Ensemble].set_param(
            'param', nengo.params.BoolParam('param', default=False))
        net.config[ens].param = True


def test_no_root_spa_network_required():
    with nengo.Network():  # nengo.Network, not spa.Network!
        state = spa.State(16)
        state2 = spa.State(16)
        # Next line will raise an exception if the two states do not share
        # the same vocab. They are supposed to share the vocab despite not
        # being in a spa.Network.
        state >> state2



import numpy as np

import nengo
from nengo.utils.numpy import rmse

import nengo_spa as spa
from nengo_spa.vocab import VocabularyMap


def test_basic():
    with spa.Network():
        bind = spa.Bind(vocab=16)

    # all inputs and outputs should share the same vocab
    vocab_a = spa.Network.get_input_vocab(bind.input_a)
    vocab_b = spa.Network.get_input_vocab(bind.input_b)
    assert vocab_a is vocab_b
    assert vocab_a.dimensions == 16
    assert vocab_b.dimensions == 16


def test_run(Simulator, seed):
    rng = np.random.RandomState(seed)
    vocab = spa.Vocabulary(32, rng=rng)
    vocab.populate('A; B')

    with spa.Network(seed=seed, vocabs=VocabularyMap([vocab])) as model:
        model.bind = spa.Bind(vocab=32)

        def inputA(t):
            if 0 <= t < 0.1:
                return 'A'
            else:
                return 'B'

        model.input = spa.Transcode(inputA, output_vocab=vocab)
        model.input >> model.bind.input_a
        spa.sym.A >> model.bind.input_b

    with model:
        p = nengo.Probe(model.bind.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.2)

    error = rmse(vocab.parse("(B*A).normalized()").v, sim.data[p][-1])
    assert error < 0.1

    error = rmse(vocab.parse("(A*A).normalized()").v, sim.data[p][100])
    assert error < 0.1



import nengo
import numpy as np
from numpy.testing import assert_almost_equal
import pytest

from nengo_spa.networks.identity_ensemble_array import IdentityEnsembleArray
from nengo_spa.pointer import Identity, SemanticPointer
from nengo_spa.testing import sp_close


@pytest.mark.parametrize('pointer', [Identity, SemanticPointer])
def test_identity_ensemble_array(Simulator, seed, rng, pointer):
    d = 64
    try:
        v = pointer(d, rng=rng).v
    except TypeError:
        v = pointer(d).v

    with nengo.Network(seed=seed) as model:
        ea = IdentityEnsembleArray(15, d, 4)
        input_v = nengo.Node(v)
        nengo.Connection(input_v, ea.input)
        p = nengo.Probe(ea.output, synapse=0.01)

    with Simulator(model) as sim:
        sim.run(0.3)

    actual = np.mean(sim.data[p][sim.trange() > 0.2], axis=0)
    assert np.abs(v[0] - actual[0]) < 0.1
    assert np.linalg.norm(v - actual) < 0.2


def test_add_output(Simulator, seed, rng, plt):
    d = 8
    pointer = SemanticPointer(d, rng=rng)

    with nengo.Network(seed=seed) as model:
        ea = IdentityEnsembleArray(15, d, 4)
        input_node = nengo.Node(pointer.v)
        nengo.Connection(input_node, ea.input)
        out = ea.add_output('const', lambda x: -x)
        assert ea.const is out
        p = nengo.Probe(out, synapse=0.01)

    with Simulator(model) as sim:
        sim.run(0.3)

    plt.plot(sim.trange(), np.dot(sim.data[p], -pointer.v))
    assert sp_close(sim.trange(), sim.data[p], -pointer, skip=0.2, atol=0.3)


def test_add_output_multiple_fn(Simulator, seed, rng, plt):
    d = 8
    pointer = SemanticPointer(d, rng=rng)

    with nengo.Network(seed=seed) as model:
        ea = IdentityEnsembleArray(15, d, 4)
        input_node = nengo.Node(pointer.v)
        nengo.Connection(input_node, ea.input)
        out = ea.add_output(
            'const', (lambda x: -x, lambda x: .5 * x, lambda x: x))
        assert ea.const is out
        p = nengo.Probe(out, synapse=0.01)

    with Simulator(model) as sim:
        sim.run(0.3)

    v = np.array(pointer.v)
    v[0] *= -1.
    v[1:4] *= .5
    expected = SemanticPointer(v)

    plt.plot(sim.trange(), np.dot(sim.data[p], expected.v))
    assert sp_close(sim.trange(), sim.data[p], expected, skip=0.2, atol=0.3)


def test_neuron_connections(Simulator, seed, rng):
    d = 8
    pointer = SemanticPointer(d, rng=rng)

    with nengo.Network(seed=seed) as model:
        ea = IdentityEnsembleArray(15, d, 4)
        input_node = nengo.Node(pointer.v)
        nengo.Connection(input_node, ea.input)

        bias = nengo.Node(1)
        neuron_in = ea.add_neuron_input()
        assert ea.neuron_input is neuron_in
        nengo.Connection(
            bias, neuron_in, transform=-3. * np.ones((neuron_in.size_in, 1)))

        neuron_out = ea.add_neuron_output()
        assert ea.neuron_output is neuron_out
        p = nengo.Probe(neuron_out)

    with Simulator(model) as sim:
        sim.run(0.3)

    assert_almost_equal(sim.data[p][sim.trange() > 0.1], 0.)



"""AST classes for dynamic operations (i.e. their output changes over time)."""

import weakref

import nengo
from nengo.utils.compat import is_number
import numpy as np

from nengo_spa.ast.base import infer_types, Node, TypeCheckedBinaryOp
from nengo_spa.ast.symbolic import FixedScalar, PointerSymbol, Symbol
from nengo_spa.exceptions import SpaTypeError
from nengo_spa.types import TAnyVocab, TScalar, TAnyVocabOfDim, TVocabulary


BasalGangliaRealization = None
BindRealization = None
DotProductRealization = None
ProductRealization = None
ScalarRealization = None
StateRealization = None
ThalamusRealization = None


input_network_registry = weakref.WeakKeyDictionary()
input_vocab_registry = weakref.WeakKeyDictionary()
output_vocab_registry = weakref.WeakKeyDictionary()


def as_node(obj):
    if is_number(obj):
        obj = FixedScalar(obj)
    return obj


binary_node_op = TypeCheckedBinaryOp(Node, as_node)


class DynamicNode(Node):
    """Base class for AST node with an output that changes over time."""

    def __invert__(self):
        if not hasattr(self.type, 'dimensions'):
            raise SpaTypeError(
                "Cannot invert semantic pointer of unknown dimensionality.")
        dimensions = self.type.dimensions
        transform = np.eye(dimensions)[-np.arange(dimensions)]
        return Transformed(self, transform, self.type)

    def __neg__(self):
        return Transformed(self, transform=-1, type_=self.type)

    @binary_node_op
    def __add__(self, other):
        type_ = infer_types(self, other)
        return Summed((self, other), type_)

    @binary_node_op
    def __radd__(self, other):
        return self + other

    @binary_node_op
    def __sub__(self, other):
        return self + (-other)

    @binary_node_op
    def __rsub__(self, other):
        return (-self) + other

    def _mul_with_fixed(self, other):
        infer_types(self, other)
        if other.type == TScalar:
            tr = other.value
        elif self.type == TScalar and other.type == TAnyVocab:
            raise SpaTypeError(
                "Cannot infer vocabulary for fixed pointer when multiplying "
                "with scalar.")
        elif isinstance(other.type, TVocabulary):
            if self.type == TScalar:
                tr = other.evaluate().v
            else:
                tr = other.evaluate().get_convolution_matrix()
        else:
            raise AssertionError("Unexpected node type in multiply.")
        return Transformed(self, tr, self.type)

    def _mul_with_dynamic(self, other, swap_inputs=False):
        type_ = infer_types(self, other)
        if type_ == TScalar:
            mul = ProductRealization()
        elif self.type == TScalar or other.type == TScalar:
            raise NotImplementedError(
                "Dynamic scaling of semantic pointer not implemented.")
        else:
            mul = BindRealization(self.type.vocab)

        if swap_inputs:
            a, b = other, self
        else:
            a, b = self, other
        a.connect_to(mul.input_a)
        b.connect_to(mul.input_b)
        return ModuleOutput(mul.output, type_)

    @binary_node_op
    def __mul__(self, other):
        if isinstance(other, Symbol):
            return self._mul_with_fixed(other)
        else:
            return self._mul_with_dynamic(other)

    @binary_node_op
    def __rmul__(self, other):
        if isinstance(other, Symbol):
            return self._mul_with_fixed(other)
        else:
            return self._mul_with_dynamic(other, swap_inputs=True)

    @binary_node_op
    def dot(self, other):
        type_ = infer_types(self, other)

        if self.type == TScalar or other.type == TScalar:
            raise SpaTypeError("Cannot do a dot product with a scalar.")

        if isinstance(other, PointerSymbol):
            tr = np.atleast_2d(other.evaluate().v)
            return Transformed(self, tr, TScalar)
        else:
            net = DotProductRealization(type_.vocab)
            self.connect_to(net.input_a)
            other.connect_to(net.input_b)
            return ModuleOutput(net.output, TScalar)

    def __matmul__(self, other):
        return self.dot(other)

    def rdot(self, other):
        return self.dot(other)

    def __rmatmul__(self, other):
        return self.rdot(other)

    def reinterpret(self, vocab=None):
        return Transformed(
            self, np.eye(self.type.dimensions),
            TAnyVocabOfDim(self.type.dimensions)
            if vocab is None else TVocabulary(vocab))

    def translate(self, vocab, populate=None, keys=None, solver=None):
        tr = self.type.vocab.transform_to(vocab, populate, keys, solver)
        return Transformed(self, tr, TVocabulary(vocab))


class Transformed(DynamicNode):
    """AST node representing a transform.

    Parameters
    ----------
    source : NengoObject
        Nengo object providing the output to apply the transform to.
    transform : ndarray
        Transform to apply.
    type_ : nengo_spa.types.Type
        The resulting type.
    """

    def __init__(self, source, transform, type_):
        super(Transformed, self).__init__(type_=type_)
        self.source = source
        self.transform = transform

    def connect_to(self, sink, **kwargs):
        if 'transform' in kwargs:
            transform = np.dot(kwargs.pop('transform'), self.transform)
        else:
            transform = self.transform
        return self.source.connect_to(sink, transform=transform, **kwargs)

    def construct(self):
        if self.type == TScalar:
            size_in = 1
        else:
            size_in = self.type.vocab.dimensions
        node = nengo.Node(size_in=size_in)
        self.connect_to(node, synapse=None)
        return node


class Summed(DynamicNode):
    """AST node representing the sum of multiple nodes.

    Parameters
    ----------
    sources : sequence of NengoObject
        Sequence of Nengo objects providing outputs to be summed.
    type_ : nengo_spa.types.Type
        The resulting type.
    """

    def __init__(self, sources, type_):
        super(Summed, self).__init__(type_=type_)
        self.sources = sources

    def connect_to(self, sink, **kwargs):
        for s in self.sources:
            s.connect_to(sink, **kwargs)

    def construct(self):
        dimensions = 1 if self.type == TScalar else self.type.dimensions
        node = nengo.Node(size_in=dimensions)
        self.connect_to(node, synapse=None)
        return node


class ModuleOutput(DynamicNode):
    """AST node representing the output of a SPA module.

    Parameters
    ----------
    output : NengoObject
        Nengo object providing the module output.
    type_ : nengo_spa.types.Type
        The resulting type.
    """

    def __init__(self, output, type_):
        super(ModuleOutput, self).__init__(type_=type_)
        self.output = output

    def construct(self):
        return self.output

    def connect_to(self, sink, **kwargs):
        nengo.Connection(self.output, sink, **kwargs)



import nengo
from nengo.exceptions import ValidationError
import numpy as np
from numpy.testing import assert_almost_equal
import pytest

import nengo_spa as spa
from nengo_spa.modules.transcode import Transcode
from nengo_spa.testing import sp_close


def test_fixed(Simulator, seed):
    with spa.Network(seed=seed) as model:
        model.buffer1 = spa.State(vocab=16)
        model.buffer2 = spa.State(vocab=8, subdimensions=8)
        model.input1 = spa.Transcode('A', output_vocab=16)
        model.input2 = spa.Transcode('B', output_vocab=8)
        model.input1 >> model.buffer1
        model.input2 >> model.buffer2
        p1 = nengo.Probe(model.buffer1.output, synapse=0.03)
        p2 = nengo.Probe(model.buffer2.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.1)

    assert sp_close(sim.trange(), sim.data[p1], model.buffer1.vocab.parse('A'),
                    skip=0.08)
    assert sp_close(sim.trange(), sim.data[p2], model.buffer2.vocab.parse('B'),
                    skip=0.08)


def test_time_varying_encode(Simulator, seed):
    with spa.Network(seed=seed) as model:
        model.buffer = spa.State(vocab=16)

        def stimulus(t):
            if t < 0.1:
                return 'A'
            elif t < 0.2:
                return model.buffer.vocab.parse('B')
            elif t < 0.3:
                return model.buffer.vocab.parse('C').v
            else:
                return '0'

        model.encode = spa.Transcode(stimulus, output_vocab=16)
        model.encode >> model.buffer

        p = nengo.Probe(model.buffer.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.3)

    vocab = model.buffer.vocab

    assert sp_close(
        sim.trange(), sim.data[p], vocab.parse('A'), skip=0.08, duration=0.02)
    assert sp_close(
        sim.trange(), sim.data[p], vocab.parse('B'), skip=0.18, duration=0.02)
    assert sp_close(
        sim.trange(), sim.data[p], vocab.parse('C'), skip=0.28, duration=0.02)
    assert sp_close(
        sim.trange(), sim.data[p], vocab.parse('0'), skip=0.38, duration=0.02)


def test_encode_with_input(Simulator, seed):
    with spa.Network(seed=seed) as model:
        buffer = spa.State(vocab=16)

        def stimulus(t, x):
            return x[0] * buffer.vocab.parse('A')

        ctrl = nengo.Node(lambda t: t > 0.2)

        encode = spa.Transcode(stimulus, output_vocab=16, size_in=1)
        nengo.Connection(ctrl, encode.input)
        encode >> buffer

        p = nengo.Probe(buffer.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.4)

    vocab = buffer.vocab
    assert sp_close(sim.trange(), sim.data[p], vocab.parse('0'), duration=0.2)
    assert sp_close(
        sim.trange(), sim.data[p], vocab.parse('A'), skip=.38, duration=0.02)


def test_transcode(Simulator, seed):
    def transcode_fn(t, sp):
        assert t < 0.15 or sp.vocab.parse('A').dot(sp) > 0.8
        return 'B'

    with spa.Network(seed=seed) as model:
        transcode = Transcode(
            transcode_fn, input_vocab=16, output_vocab=16)
        spa.sym.A >> transcode
        p = nengo.Probe(transcode.output, synapse=None)

    with Simulator(model) as sim:
        sim.run(0.2)

    assert sp_close(sim.trange(), sim.data[p],
                    transcode.output_vocab.parse('B'))


def test_passthrough(Simulator, seed):
    with spa.Network(seed=seed) as model:
        passthrough = Transcode(input_vocab=16, output_vocab=16)
        spa.sym.A >> passthrough
        p = nengo.Probe(passthrough.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.2)

    assert sp_close(sim.trange(), sim.data[p],
                    passthrough.output_vocab.parse('A'), skip=0.18)


def test_decode(Simulator, seed):
    class OutputFn(object):
        def __init__(self):
            self.called = False

        def __call__(self, t, v):
            if t > 0.001:
                self.called = True
                assert_almost_equal(v.vocab.parse('A').v, v.v)

    output_fn = OutputFn()

    with spa.Network(seed=seed) as model:
        model.config[nengo.Connection].synapse = nengo.Lowpass(0.)
        model.output = Transcode(output_fn, input_vocab=16)
        spa.sym.A >> model.output

    with Simulator(model) as sim:
        sim.run(0.01)

    assert output_fn.called


def test_decode_with_output(Simulator, seed):
    def decode_fn(t, v):
        return [t]

    with spa.Network(seed=seed) as model:
        model.decode = Transcode(decode_fn, input_vocab=16)
        p = nengo.Probe(model.decode.output)

    with Simulator(model) as sim:
        sim.run(0.01)

    assert_almost_equal(np.squeeze(sim.data[p]), sim.trange())


def test_decode_size_out(Simulator, seed):
    def decode_fn(t, v):
        return [t]

    with spa.Network(seed=seed) as model:
        model.decode = Transcode(decode_fn, input_vocab=16, size_out=1)
        p = nengo.Probe(model.decode.output)

    with Simulator(model) as sim:
        sim.run(0.01)

    assert_almost_equal(np.squeeze(sim.data[p]), sim.trange())


def test_exception_when_no_vocabularies_are_given():
    with spa.Network():
        with pytest.raises(ValidationError):
            Transcode('A')
        with pytest.raises(ValidationError):
            Transcode(lambda t: 'A')


@pytest.mark.parametrize('value', [
    'String',
    spa.pointer.Zero(32),
    spa.sym.Symbol,
    lambda t: 'String',
    lambda t: spa.pointer.Zero(32),
    lambda t: spa.sym.Symbol,
])
def test_output_types(Simulator, value):
    with spa.Network() as model:
        stim = spa.Transcode(value, output_vocab=32)
        state = spa.State(32)
        stim >> state

    with Simulator(model) as sim:
        sim.run(0.01)



from nengo.params import BoolParam, Default, IntParam

import nengo_spa
from nengo_spa.network import Network
from nengo_spa.vocab import VocabularyOrDimParam


class Bind(Network):
    """Network for binding together two inputs.

    Binding is done with circular convolution. For more details on how
    this is computed, see the underlying `.CircularConvolution`
    network.

    Parameters
    ----------
    vocab : Vocabulary or int
        The vocabulary to use to interpret the vector. If an integer is given,
        the default vocabulary of that dimensionality will be used.
    neurons_per_dimension : int, optional (Default: 200)
        Number of neurons to use in each product computation.
    invert_a : bool, optional
        Whether to reverse the order of elements in first input.
    invert_b : bool, optional
        Whether to reverse the order of elements in the second input.
        Flipping exactly one input will make the network perform circular
        correlation instead of circular convolution which can be treated as an
        approximate inverse to circular convolution.
    kwargs : dict
        Keyword arguments passed through to `nengo_spa.Network`.

    Attributes
    ----------
    input_a : nengo.Node
        First input vector.
    input_b : nengo.Node
        Second input vector.
    output : nengo.Node
        Output.
    """

    vocab = VocabularyOrDimParam('vocab', default=None, readonly=True)
    neurons_per_dimension = IntParam(
        'neurons_per_dimension', default=200, low=1, readonly=True)
    invert_a = BoolParam('invert_a', default=False, readonly=True)
    invert_b = BoolParam('invert_b', default=False, readonly=True)

    def __init__(self, vocab=Default, neurons_per_dimension=Default,
                 invert_a=Default, invert_b=Default, **kwargs):
        kwargs.setdefault('label', "Bind")
        super(Bind, self).__init__(**kwargs)

        self.vocab = vocab
        self.neurons_per_dimension = neurons_per_dimension
        self.invert_a = invert_a
        self.invert_b = invert_b

        with self:
            self.cc = nengo_spa.networks.CircularConvolution(
                self.neurons_per_dimension, self.vocab.dimensions,
                self.invert_a, self.invert_b)

        self.input_a = self.cc.input_a
        self.input_b = self.cc.input_b
        self.output = self.cc.output

        self.declare_input(self.input_a, self.vocab)
        self.declare_input(self.input_b, self.vocab)
        self.declare_output(self.output, self.vocab)



"""Operator-like SPA functions.

The operator functions work similar to other overloaded operators in Python.
They call the class implementation on the operand, e.g. ``reinterpret(a, ...)``
calls ``a.reinterpret(...)`` like ``-a`` calls ``a.__neg__()``. For binary
operators, this will be tried for the left operand first. If the left operand
does not have the corresponding method or if it returns `NotImplemented`, the
same method prefixed with an ``r`` will be called on the right operand. For
example, ``dot(a, b)`` tries ``a.dot(b)`` first and then ``b.rdot(a)``. This is
equivalent to for example ``a + b`` trying ``a.__add__(b)`` first and then
``b.__radd__(a)``.
"""


def dot(a, b):
    """Dot-product between *a* and *b*."""
    result = NotImplemented
    if hasattr(a, 'dot'):
        result = a.dot(b)
    if result is NotImplemented and hasattr(b, 'dot'):
        result = b.rdot(a)
    if result is NotImplemented:
        raise TypeError(
            "'dot' not supported between instances of '{}' and '{}'".format(
                type(a), type(b)))
    return result


def reinterpret(source, vocab=None):
    """Reinterpret *source* Semantic Pointer as part of vocabulary *vocab*.

    The *vocab* parameter can be set to *None* to clear the associated
    vocabulary and allow the *source* to be interpreted as part of the
    vocabulary of any Semantic Pointer it is combined with.
    """
    if hasattr(source, 'reinterpret'):
        return source.reinterpret(vocab)
    else:
        raise TypeError("bad operand type for 'reinterpret'")


def translate(source, vocab, populate=None, keys=None, solver=None):
    """Translate *source* Semantic Pointer to vocabulary *vocab*.

    The translation of a Semantic Pointer uses some form of projection to
    convert the source Semantic Pointer to a Semantic Pointer of another
    vocabulary. By default the outer products of terms in the source and
    target vocabulary are used, but if *solver* is given, it is used to find
    a least squares solution for this projection.

    Parameters
    ----------
    source : object
        Source of Semantic Pointer.
    vocab : Vocabulary
        Target vocabulary.
    populate : bool, optional
        Whether the target vocabulary should be populated with missing keys.
        This is done by default, but with a warning. Set this explicitly to
        *True* or *False* to silence the warning or raise an error.
    keys : list, optional
        All keys to translate. If *None*, all keys in the source vocabulary
        will be translated.
    solver : nengo.Solver, optional
        If given, the solver will be used to solve the least squares problem to
        provide a better projection for the translation.
    """
    if hasattr(source, 'translate'):
        return source.translate(vocab, populate, keys, solver)
    else:
        raise TypeError("bad operand type for 'translate'")



#!/usr/bin/env python
import imp
import io
import os
import sys

try:
    from setuptools import find_packages, setup
except ImportError:
    raise ImportError(
        "'setuptools' is required but not installed. To install it, "
        "follow the instructions at "
        "https://pip.pypa.io/en/stable/installing/#installing-with-get-pip-py")


def read(*filenames, **kwargs):
    encoding = kwargs.get("encoding", "utf-8")
    sep = kwargs.get("sep", "\n")
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


root = os.path.dirname(os.path.realpath(__file__))
version_module = imp.load_source(
    "version", os.path.join(root, "nengo_spa", "version.py"))
testing = "test" in sys.argv or "pytest" in sys.argv

docs_require = [
    "jupyter_client",
    "sphinx",
    "nbsphinx",
    "sphinx_rtd_theme",
]
optional_requires = [
    "scipy"
]
tests_require = [
    "jupyter",
    "matplotlib>=2.0",
    "pytest>=2.3",
]

setup(
    name="nengo_spa",
    version=version_module.version,
    author="Applied Brain Research",
    author_email="info@appliedbrainresearch.com",
    url="https://www.nengo.ai/nengo-spa/",
    packages=find_packages(exclude=["*.tests"]),
    scripts=[],
    license="Free for non-commercial use",
    description="An implementation of the Semantic Pointer Architecture for "
                "Nengo",
    long_description=read("README.rst", "CHANGES.rst"),
    zip_safe=True,
    include_package_data=True,
    setup_requires=["pytest-runner"] if testing else [] + [
        "numpy>=1.8",
    ],
    install_requires = [
        "nengo>=2.7",
        "numpy>=1.8",
    ],
    extras_require = {
        "all": docs_require + optional_requires + tests_require,
        "docs": docs_require,
        "optional": optional_requires,
        "tests": tests_require
    },
    tests_require=tests_require,
    entry_points={
    },
    classifiers=[  # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: Free for non-commercial use",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



import pytest

import nengo_spa as spa
from nengo_spa.exceptions import SpaParseError, SpaTypeError


def test_missing_source():
    d = 16
    with spa.Network():
        a = spa.State(d)
        assert a
        with pytest.raises(NameError):
            b >> a  # noqa: F821


@pytest.mark.parametrize('sink', ('b', 'B'))
def test_missing_sink(sink):
    d = 16
    with spa.Network():
        a = spa.State(d)
        assert a
        with pytest.raises(NameError):
            eval('a >> %s' % sink)


@pytest.mark.parametrize('sink', ('0.5', '0.6 * a', 'a * 0.6'))
def test_invalid_sink(sink):
    d = 16
    with spa.Network():
        a = spa.State(d)
        assert a
        with pytest.raises((SyntaxError, SpaTypeError)):
            eval('a >> {}'.format(sink))


def test_missing_pointer():
    vocab = spa.Vocabulary(16)
    with spa.Network():
        a = spa.State(vocab)
        assert a
        with pytest.raises(SpaParseError):
            spa.sym.A >> a



"""Functions to examine data and vectors produced by SPA."""

from itertools import combinations

import numpy as np

import nengo.utils.numpy as npext
from nengo.exceptions import ValidationError
from nengo.utils.compat import is_iterable

from nengo_spa.pointer import SemanticPointer
from nengo_spa.vocab import Vocabulary


def similarity(data, vocab, normalize=False):
    """Return the similarity between simulation data and Semantic Pointers.

    Computes the dot products between all Semantic Pointers in the Vocabulary
    and the simulation data for each timestep. If ``normalize=True``,
    normalizes all vectors to compute the cosine similarity.

    Parameters
    ----------
    data: (D,) or (T, D) array_like
        The *D*-dimensional data for *T* timesteps used for comparison.
    vocab: Vocabulary or array_like
        Vocabulary (or list of vectors) used to calculate the similarity
        values.
    normalize : bool, optional
        Whether to normalize all vectors, to compute the cosine similarity.
    """

    if isinstance(data, SemanticPointer):
        data = data.v

    if isinstance(vocab, Vocabulary):
        vectors = vocab.vectors
    elif is_iterable(vocab):
        if isinstance(next(iter(vocab)), SemanticPointer):
            vocab = [p.v for p in vocab]
        vectors = np.array(vocab, copy=False, ndmin=2)
    else:
        raise ValidationError("%r object is not a valid vocabulary"
                              % (type(vocab).__name__), attr='vocab')

    dots = np.dot(vectors, data.T)

    if normalize:
        # Zero-norm vectors should return zero, so avoid divide-by-zero error
        eps = np.nextafter(0, 1)  # smallest float above zero
        dnorm = np.maximum(npext.norm(data.T, axis=0, keepdims=True), eps)
        vnorm = np.maximum(npext.norm(vectors, axis=1, keepdims=True), eps)

        if len(dots.shape) == 1:
            vnorm = np.squeeze(vnorm)

        dots /= dnorm
        dots /= vnorm

    return dots.T


def pairs(vocab):
    """Return expressions for all possible combinations to bind *vocab*'s keys.

    Examples
    --------

    >>> vocab = nengo_spa.Vocabulary(32)
    >>> vocab.populate('A; B; C')
    >>> sorted(pairs(vocab))
    ['A*B', 'A*C', 'B*C']
    """

    return set(x + '*' + y for x, y in combinations(vocab.keys(), 2))


def text(v, vocab, minimum_count=1, maximum_count=None,
         threshold=0.1, join=';', terms=None, normalize=False):
    """Return a human-readable text version of the provided vector.

    This is meant to give a quick text version of a vector for display
    purposes. To do this, compute the dot product between the vector
    and all the terms in the vocabulary. The top few vectors are
    chosen for inclusion in the text. It will try to only return
    terms with a match above the *threshold*, but will always return
    at least *minimum_count* and at most maximum_count terms. Terms
    are sorted from most to least similar.

    Parameters
    ----------
    v : SemanticPointer or array_like
        The vector to convert into text.
    minimum_count : int, optional
        Always return at least this many terms in the text.
    maximum_count : int, optional
        Never return more than this many terms in the text.
        If None, all terms will be returned.
    threshold : float, optional
        How small a similarity for a term to be ignored.
    join : str, optional
        The text separator to use between terms.
    terms : list, optional
        Only consider terms in this list of strings.
    normalize : bool
        Whether to normalize the vector before computing similarity.
    """
    if not isinstance(v, SemanticPointer):
        v = SemanticPointer(v)
    if normalize:
        v = v.normalized()

    if terms is None:
        terms = vocab.keys()
        vectors = vocab.vectors
    else:
        vectors = vocab.parse_n(*terms)

    matches = list(zip(similarity(v, vectors), terms))
    matches.sort()
    matches.reverse()

    r = []
    for m in matches:
        if minimum_count is not None and len(r) < minimum_count:
            r.append(m)
        elif maximum_count is not None and len(r) == maximum_count:
            break
        elif threshold is None or m[0] > threshold:
            r.append(m)
        else:
            break

    return join.join(['%0.2f%s' % (sim, key) for (sim, key) in r])



import nengo
import numpy as np
from numpy.testing import assert_allclose
import pytest

import nengo_spa as spa
from nengo_spa.actions import ActionSelection
from nengo_spa.ast.symbolic import PointerSymbol
from nengo_spa.exceptions import SpaActionSelectionError, SpaTypeError
from nengo_spa.testing import sp_close


def test_new_action_syntax(Simulator, seed, plt, rng):
    model = spa.Network(seed=seed)
    model.config[spa.State].vocab = 3
    model.config[spa.State].subdimensions = 3
    with model:
        model.ctrl = spa.State(16, subdimensions=16, label='ctrl')

        def input_func(t):
            if t < 0.2:
                return 'A'
            elif t < 0.4:
                return 'B'
            else:
                return 'C'

        model.input = spa.Transcode(input_func, output_vocab=16)

        model.state = spa.State(label='state')
        model.buff1 = spa.State(label='buff1')
        model.buff2 = spa.State(label='buff2')
        model.buff3 = spa.State(label='buff3')

        node1 = nengo.Node([0, 1, 0])
        node2 = nengo.Node([0, 0, 1])

        nengo.Connection(node1, model.buff1.input)
        nengo.Connection(node2, model.buff2.input)

        model.input >> model.ctrl
        model.buff1 >> model.state
        with spa.ActionSelection():
            spa.ifmax(
                spa.dot(model.ctrl, spa.sym.A), model.buff1 >> model.buff3)
            spa.ifmax(
                spa.dot(model.ctrl, spa.sym.B), model.buff2 >> model.buff3)
            spa.ifmax(
                spa.dot(model.ctrl, spa.sym.C),
                model.buff1 * model.buff2 >> model.buff3)

        state_probe = nengo.Probe(model.state.output, synapse=0.03)
        buff3_probe = nengo.Probe(model.buff3.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.6)

    data = sim.data[buff3_probe]

    plt.plot(sim.trange(), data)

    state_val = np.mean(sim.data[state_probe][200:], axis=0)
    assert state_val[0] < 0.2
    assert state_val[1] > 0.8
    assert state_val[2] < 0.2

    valueA = np.mean(data[150:200], axis=0)  # should be [0, 1, 0]
    valueB = np.mean(data[350:400], axis=0)  # should be [0, 0, 1]
    valueC = np.mean(data[550:600], axis=0)  # should be [1, 0, 0]

    assert valueA[0] < 0.2
    assert valueA[1] > 0.75
    assert valueA[2] < 0.2

    assert valueB[0] < 0.2
    assert valueB[1] < 0.2
    assert valueB[2] > 0.75

    assert valueC[0] > 0.75
    assert valueC[1] < 0.2
    assert valueC[2] < 0.2


def test_dummy_action():
    with spa.Network():
        with spa.ActionSelection():
            spa.ifmax(0)
            spa.ifmax('named-dummy', 0)

            with pytest.raises(ValueError):
                spa.ifmax('must-provide-condition')


def test_dot_product(Simulator, seed, plt):
    d = 16

    with spa.Network(seed=seed) as model:
        model.state_a = spa.State(d)
        model.state_b = spa.State(d)
        model.result = spa.Scalar()

        model.stimulus = spa.Transcode(
            lambda t: 'A' if t <= 0.3 else 'B', output_vocab=d)

        spa.sym.A >> model.state_a
        model.stimulus >> model.state_b
        spa.dot(model.state_a, model.state_b) >> model.result
        p = nengo.Probe(model.result.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.6)

    plt.plot(sim.trange(), sim.data[p])

    t = sim.trange()
    assert np.mean(sim.data[p][(t > 0.1) & (t <= 0.3)]) > 0.8
    assert np.mean(sim.data[p][t > 0.4]) < 0.15


class TestExceptions():
    @pytest.fixture
    def model(self):
        with spa.Network() as model:
            model.state_a = spa.State(16)
            model.state_b = spa.State(32)
            model.state_c = spa.State(32)
            model.scalar = spa.Scalar()
        return model

    def test_invalid_types_in_binary_operation(self, model):
        with model:
            with pytest.raises(SpaTypeError):
                model.state_a + model.state_b >> model.state_c

    def test_approx_inverse_of_scalar(self, model):
        with model:
            with pytest.raises(SpaTypeError):
                ~model.scalar >> model.state_c

    def test_dot_product_incompatiple_vocabs(self, model):
        with model:
            with pytest.raises(SpaTypeError):
                spa.dot(model.state_a, model.state_b) >> model.scalar

    def test_dot_product_first_argument_invalid(self, model):
        with model:
            with pytest.raises(SpaTypeError):
                spa.dot(model.scalar, model.state_b) >> model.scalar

    def test_dot_product_second_argument_invalid(self, model):
        with model:
            with pytest.raises(SpaTypeError):
                spa.dot(model.state_a, model.scalar) >> model.scalar

    def test_cast_type_inference_not_possible(self, model):
        with model:
            with pytest.raises(SpaTypeError):
                spa.dot(
                    spa.reinterpret(model.state_a), spa.sym.A) >> model.scalar

    def test_reinterpret_non_matching_dimensions(self, model):
        with model:
            with pytest.raises(SpaTypeError):
                spa.reinterpret(model.state_b) >> model.state_a


def test_access_thal_and_bg_objects():
    with spa.Network() as m:
        m.a = spa.Scalar()
        m.b = spa.Scalar()

        m.c = spa.Scalar()
        m.d = spa.Scalar()

        with spa.ActionSelection() as actions:
            spa.ifmax(m.a, 0 >> m.c)
            spa.ifmax(m.b, 1 >> m.c)

    assert isinstance(actions.bg, spa.BasalGanglia)
    assert isinstance(actions.thalamus, spa.Thalamus)


def test_eval(Simulator):
    with spa.Network() as net:
        a = spa.Transcode(input_vocab=16)
        0.5 * spa.sym.A >> a
        p = nengo.Probe(a.output)

    with Simulator(net) as sim:
        sim.run(1.0)

    assert np.allclose(sim.data[p][-1], net.vocabs[16].parse("0.5*A").v)


def test_assignment_of_fixed_scalar(Simulator, rng):
    with spa.Network() as model:
        sink = spa.Scalar()
        0.5 >> sink
        p = nengo.Probe(sink.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    assert_allclose(sim.data[p][sim.trange() > 0.3], 0.5, atol=0.2)


def test_assignment_of_pointer_symbol(Simulator, rng):
    vocab = spa.Vocabulary(16, rng=rng)
    vocab.populate('A')

    with spa.Network() as model:
        sink = spa.State(vocab)
        PointerSymbol('A') >> sink
        p = nengo.Probe(sink.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    assert sp_close(sim.trange(), sim.data[p], vocab['A'], skip=0.3)


def test_assignment_of_dynamic_scalar(Simulator, rng):
    with spa.Network() as model:
        source = spa.Scalar()
        sink = spa.Scalar()
        nengo.Connection(nengo.Node(0.5), source.input)
        source >> sink
        p = nengo.Probe(sink.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    assert_allclose(sim.data[p][sim.trange() > 0.3], 0.5, atol=0.2)


@pytest.mark.parametrize('source,use_ens', [
    ('nengo.Node(0.5)', False),
    ('nengo.Node([0., 0.5, 1.])[1]', False),
    ('nengo.Node(0.5)', True)])
def test_assignment_of_nengo_scalar(source, use_ens, Simulator, rng):
    with spa.Network() as model:
        source = eval(source)
        if use_ens:
            ens = nengo.Ensemble(50, 1)
            nengo.Connection(source, ens)
            source = ens
        sink = spa.Scalar()
        source >> sink
        p = nengo.Probe(sink.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    assert_allclose(sim.data[p][sim.trange() > 0.3], 0.5, atol=0.2)


def test_assignment_of_dynamic_pointer(Simulator, rng):
    vocab = spa.Vocabulary(16, rng=rng)
    vocab.populate('A')

    with spa.Network() as model:
        source = spa.Transcode('A', output_vocab=vocab)
        sink = spa.State(vocab)
        source >> sink
        p = nengo.Probe(sink.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    assert sp_close(sim.trange(), sim.data[p], vocab['A'], skip=0.3)


def test_non_default_input_and_output(Simulator, rng):
    vocab = spa.Vocabulary(32, rng=rng)
    vocab.populate('A; B')

    with spa.Network() as model:
        a = spa.Transcode('A', output_vocab=vocab)
        b = spa.Transcode('B', output_vocab=vocab)
        bind = spa.Bind(vocab)
        a.output >> bind.input_a
        b.output >> bind.input_b
        p = nengo.Probe(bind.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    assert sp_close(
        sim.trange(), sim.data[p], vocab.parse('A*B'), skip=0.3, atol=0.3)


def test_action_selection(Simulator, rng):
    vocab = spa.Vocabulary(64)
    vocab.populate('A; B; C; D; E; F')

    with spa.Network() as model:
        state = spa.Transcode(
            lambda t: 'ABCDEF'[min(5, int(t / 0.5))], output_vocab=vocab)
        scalar = spa.Scalar()
        pointer = spa.State(vocab)
        with ActionSelection():
            spa.ifmax(spa.dot(state, PointerSymbol('A')), 0.5 >> scalar)
            spa.ifmax(
                spa.dot(state, PointerSymbol('B')),
                PointerSymbol('B') >> pointer)
            spa.ifmax(spa.dot(state, PointerSymbol('C')), state >> pointer)
            d_utility = spa.ifmax(0, PointerSymbol('D') >> pointer)
            spa.ifmax(
                spa.dot(state, PointerSymbol('E')),
                0.25 >> scalar, PointerSymbol('E') >> pointer)
        nengo.Connection(
            nengo.Node(lambda t: 1.5 < t <= 2.), d_utility)
        p_scalar = nengo.Probe(scalar.output, synapse=0.03)
        p_pointer = nengo.Probe(pointer.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(3.)

    t = sim.trange()
    assert_allclose(sim.data[p_scalar][(0.3 < t) & (t <= 0.5)], 0.5, atol=0.2)
    assert sp_close(
        sim.trange(), sim.data[p_pointer], vocab['B'], skip=0.8, duration=0.2)
    assert sp_close(
        sim.trange(), sim.data[p_pointer], vocab['C'], skip=1.3, duration=0.2)
    assert sp_close(
        sim.trange(), sim.data[p_pointer], vocab['D'], skip=1.8, duration=0.2)
    assert_allclose(sim.data[p_scalar][(2.3 < t) & (t <= 2.5)], 0.25, atol=0.2)
    assert sp_close(
        sim.trange(), sim.data[p_pointer], vocab['E'], skip=2.3, duration=0.2)


def test_does_not_allow_nesting_of_action_selection():
    with spa.Network():
        with ActionSelection():
            with pytest.raises(SpaActionSelectionError):
                with ActionSelection():
                    pass


def test_action_selection_enforces_connections_to_be_part_of_action():
    with spa.Network():
        state1 = spa.State(16)
        state2 = spa.State(16)
        with pytest.raises(SpaActionSelectionError):
            with ActionSelection():
                    state1 >> state2


def test_action_selection_is_not_built_on_exception():
    with spa.Network():
        state1 = spa.State(16)
        state2 = spa.State(32)
        with pytest.raises(SpaTypeError):
            with ActionSelection() as action_sel:
                spa.ifmax(1., state1 >> state2)
    assert not action_sel.built


def test_naming_of_actions():
    with spa.Network():
        state1 = spa.State(16)
        state2 = spa.State(16)
        with ActionSelection() as action_sel:
            u0 = spa.ifmax('name0', 0., state1 >> state2)
            u1 = spa.ifmax(0., state1 >> state2)
            u2 = spa.ifmax('name2', 0., state1 >> state2)

    assert tuple(action_sel.keys()) == ('name0', 1, 'name2')
    assert action_sel['name0'] is u0
    assert action_sel['name2'] is u2
    for i, u in enumerate((u0, u1, u2)):
        assert action_sel[i] is u



import nengo
from nengo.params import Default, IntParam

from nengo_spa.network import Network


class Scalar(Network):
    """Represents a single scalar.

    Parameters
    ----------
    n_neurons : int, optional (Default: 50)
        Number of neurons to represent the scalar.
    kwargs : dict
        Keyword arguments passed through to `nengo_spa.Network`.

    Attributes
    ----------
    input : nengo.Node
         Input.
    output : nengo.Node
        Output.
    """

    n_neurons = IntParam('n_neurons', default=50, low=1, readonly=True)

    def __init__(self, n_neurons=Default, **kwargs):
        kwargs.setdefault('label', "Scalar")
        super(Scalar, self).__init__(**kwargs)

        self.n_neurons = n_neurons

        with self:
            self.scalar = nengo.Ensemble(self.n_neurons, 1)

        self.input = self.scalar
        self.output = self.scalar
        self.declare_input(self.input, None)
        self.declare_output(self.output, None)



"""Types used in the Nengo SPA type system to verify operations."""

from nengo_spa.exceptions import SpaTypeError


class Type(object):
    """Describes a type.

    Types can be compared and by default two types are considered to be equal
    when their class and name match. Subclasses are allowed to overwrite
    *__eq__*.

    Furthermore, a partial ordering can be defined over types by overwriting
    the *__gt__* method. A call to ``a.__gt__(b)`` should return *True*, if
    (and only if) *b* can be cast to the type *a*. For example, the type for an
    unspecified vocabulary can be cast to the type for a specific vocabulary.

    Note, that other comparison operators will be implemented on the
    implementation of *__gt__*.

    Parameters
    ----------
    name : str
        Name of the type.
    """

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.name)

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.__class__) ^ hash(self.name)

    def __eq__(self, other):
        if not isinstance(other, Type):
            return NotImplemented
        return self.__class__ is other.__class__ and self.name == other.name

    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        return other.__gt__(self)

    def __le__(self, other):
        return self < other or self == other

    def __gt__(self, other):
        if not isinstance(other, Type):
            return NotImplemented
        return False

    def __ge__(self, other):
        return self > other or self == other


TScalar = Type('TScalar')


class _TAnyVocab(Type):
    """Type that allows for any vocabulary."""

    def __gt__(self, other):
        if not isinstance(other, Type):
            return NotImplemented
        return other == TScalar


TAnyVocab = _TAnyVocab('TAnyVocab')


class TAnyVocabOfDim(Type):
    """Type that allows for any vocab of a given dimensionality."""

    def __init__(self, dimensions):
        super(TAnyVocabOfDim, self).__init__('TAnyVocabOfDim')
        self.dimensions = dimensions

    def __repr__(self):
        return '{}({!r})'.format(
            self.__class__.__name__, self.dimensions)

    def __str__(self):
        return '{}<{}>'.format(self.name, self.dimensions)

    def __eq__(self, other):
        if not isinstance(other, Type):
            return NotImplemented
        return (super(TAnyVocabOfDim, self).__eq__(other) and
                self.dimensions == other.dimensions)

    def __gt__(self, other):
        if not isinstance(other, Type):
            return NotImplemented
        return other <= TAnyVocab


class TVocabulary(Type):
    """Type for a specific vocabulary.

    All vocabulary types constitute a type class.
    """

    def __init__(self, vocab):
        super(TVocabulary, self).__init__('TVocabulary')
        self.vocab = vocab

    @property
    def dimensions(self):
        return self.vocab.dimensions

    def __repr__(self):
        return '{}({!r})'.format(
            self.__class__.__name__, self.vocab)

    def __str__(self):
        return '{}<{}>'.format(self.name, self.vocab)

    def __hash__(self):
        return super(TVocabulary, self).__hash__() ^ hash(self.vocab)

    def __eq__(self, other):
        if not isinstance(other, Type):
            return NotImplemented
        return (super(TVocabulary, self).__eq__(other) and
                self.vocab is other.vocab)

    def __gt__(self, other):
        if not isinstance(other, Type):
            return NotImplemented
        return other <= TAnyVocabOfDim(self.vocab.dimensions)


def coerce_types(*types):
    """Returns the most specific type in the argument list.

    If the types passed in the argument list are incompatible a `.SpaTypeError`
    will be raised.

    The specificity of a types is defined by their partial ordering implemented
    in the type classes.
    """
    type_ = max(types)
    if not all(t <= type_ for t in types):
        offender = next(iter(t for t in types if not t <= type_))
        if (hasattr(offender, 'vocab') and hasattr(type_, 'vocab') and
                offender.vocab is not type_.vocab):
            reason = "Different vocabularies"
        elif (hasattr(offender, 'dimensions') and
                hasattr(type_, 'dimensions') and
                offender.dimensions != type_.dimensions):
            reason = "Dimensionality mismatch"
        else:
            reason = "Incompatible types"
        raise SpaTypeError(reason + ": " + ", ".join(str(t) for t in types))
    return type_
